"""Helper functions for layout.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""
from __future__ import annotations

import numbers
import typing
from collections import defaultdict

import numpy as np
from gdstk import Label as _Label
from gdstk import Polygon
from numpy import cos, pi, sin
from numpy.linalg import norm
from rich.console import Console
from rich.table import Table

from gdsfactory.snap import snap_to_grid

if typing.TYPE_CHECKING:
    from gdsfactory.port import Port


def pprint_ports(ports: dict[str, Port] or list[Port]) -> None:
    """Prints ports in a rich table."""
    console = Console()
    table = Table(show_header=True, header_style="bold")
    ports_list = ports if isinstance(ports, list) else list(ports.values())
    if not ports_list:
        return
    p0 = ports_list[0]
    filtered_dict = {
        key: value for key, value in p0.to_dict().items() if value is not None
    }
    keys = filtered_dict.keys()

    for key in keys:
        table.add_column(key)

    for port in ports_list:
        port_dict = port.to_dict()
        row = [str(port_dict.get(key, "")) for key in keys]
        table.add_row(*row)

    console.print(table)


class Label(_Label):
    def __repr__(self) -> str:
        return f"Label(text={self.text!r}, origin={self.origin}, layer=({self.layer}, {self.texttype}))"


def get_polygons(
    instance,
    by_spec: bool | tuple[int, int] = False,
    depth: int | None = None,
    include_paths: bool = True,
    as_array: bool = True,
    as_shapely: bool = False,
    as_shapely_merged: bool = False,
) -> list[Polygon] | dict[tuple[int, int], list[Polygon]]:
    """Return a list of polygons in this cell.

    Args:
        by_spec: bool or layer
            If True, the return value is a dictionary with the
            polygons of each individual pair (layer, datatype), which
            are used as keys.  If set to a tuple of (layer, datatype),
            only polygons with that specification are returned.
        depth: integer or None
            If not None, defines from how many reference levels to
            retrieve polygons.  References below this level will result
            in a bounding box.  If `by_spec` is True the key will be the
            name of this cell.
        include_paths: If True, polygonal representation of paths are also included in the result.
        as_array: when as_array=false, return the Polygon objects instead.
            polygon objects have more information (especially when by_spec=False) and are faster to retrieve.
        as_shapely: returns shapely polygons.

    Returns
        out: list of array-like[N][2] or dictionary
            List containing the coordinates of the vertices of each
            polygon, or dictionary with with the list of polygons (if
            `by_spec` is True).

    Note:
        Instances of `FlexPath` and `RobustPath` are also included in
        the result by computing their polygonal boundary.
    """
    import gdsfactory as gf

    if hasattr(instance, "_cell"):
        layers = instance.get_layers()
        gdstk_instance = instance._cell

    else:
        layers = instance.parent.get_layers()
        gdstk_instance = instance._reference

    if not by_spec:
        polygons = gdstk_instance.get_polygons(depth=depth, include_paths=include_paths)

    elif by_spec is True:
        polygons = {
            layer: gdstk_instance.get_polygons(
                depth=depth,
                layer=layer[0],
                datatype=layer[1],
                include_paths=include_paths,
            )
            for layer in layers
        }

    else:
        by_spec = gf.get_layer(by_spec)
        polygons = gdstk_instance.get_polygons(
            depth=depth,
            layer=by_spec[0],
            datatype=by_spec[1],
            include_paths=include_paths,
        )

    if not as_array:
        return polygons
    elif as_shapely_merged:
        import shapely as sp

        polygons = [sp.Polygon(polygon.points) for polygon in polygons]
        p = sp.Polygon()
        for polygon in polygons:
            p = p | polygon
        return p

    elif as_shapely:
        import shapely as sp

        return [sp.Polygon(polygon.points) for polygon in polygons]

    elif by_spec is not True:
        return [polygon.points for polygon in polygons]
    layer_to_polygons = defaultdict(list)
    for layer, polygons_list in polygons.items():
        for polygon in polygons_list:
            layer_to_polygons[layer].append(polygon.points)
    return layer_to_polygons


def _parse_layer(layer):
    """Check if the variable layer is a Layer object, a 2-element list like \
    [0, 1] representing layer = 0 and datatype = 1, or just a layer number.

    Args:
        layer: int, array-like[2], or set Variable to check.

    Returns:
        (gds_layer, gds_datatype) : array-like[2]
            The layer number and datatype of the input.
    """
    if hasattr(layer, "gds_layer"):
        gds_layer, gds_datatype = layer.gds_layer, layer.gds_datatype
    elif np.shape(layer) == (2,):  # In form [3,0]
        gds_layer, gds_datatype = layer[0], layer[1]
    elif np.shape(layer) == (1,):  # In form [3]
        gds_layer, gds_datatype = layer[0], 0
    elif layer is None:
        gds_layer, gds_datatype = 0, 0
    elif isinstance(layer, numbers.Number):
        gds_layer, gds_datatype = layer, 0
    else:
        raise ValueError(
            """_parse_layer() was passed something
            that could not be interpreted as a layer: layer = %s"""
            % layer
        )
    if not isinstance(gds_layer, int):
        raise ValueError(f"invalid layer {layer}")
    if not isinstance(gds_datatype, int):
        raise ValueError(f"invalid layer {layer}")
    return (gds_layer, gds_datatype)


class _GeometryHelper:
    """Helper class for a class with functions move() and the property bbox.

    It uses that function+property to enable you to do things like check what the
    center of the bounding box is (self.center), and also to do things like move
    the bounding box such that its maximum x value is 5.2 (self.xmax = 5.2).
    """

    @property
    def center(self):
        """Returns the center of the bounding box."""
        return np.sum(self.bbox, 0) / 2

    @center.setter
    def center(self, destination) -> None:
        """Sets the center of the bounding box.

        Args:
            destination : array-like[2] Coordinates of the new bounding box center.
        """
        self.move(destination=destination, origin=self.center)

    @property
    def x(self):
        """Returns the x-coordinate of the center of the bounding box."""
        return np.sum(self.bbox, 0)[0] / 2

    @x.setter
    def x(self, destination) -> None:
        """Sets the x-coordinate of the center of the bounding box.

        Args:
            destination : int or float x-coordinate of the bbox center.
        """
        destination = (destination, self.center[1])
        self.move(destination=destination, origin=self.center, axis="x")

    @property
    def y(self):
        """Returns the y-coordinate of the center of the bounding box."""
        return np.sum(self.bbox, 0)[1] / 2

    @y.setter
    def y(self, destination) -> None:
        """Sets the y-coordinate of the center of the bounding box.

        Args:
        destination : int or float
            y-coordinate of the bbox center.
        """
        destination = (self.center[0], destination)
        self.move(destination=destination, origin=self.center, axis="y")

    @property
    def xmax(self):
        """Returns the maximum x-value of the bounding box."""
        return self.bbox[1][0]

    @xmax.setter
    def xmax(self, destination) -> None:
        """Sets the x-coordinate of the maximum edge of the bounding box.

        Args:
        destination : int or float
            x-coordinate of the maximum edge of the bbox.
        """
        self.move(destination=(destination, 0), origin=self.bbox[1], axis="x")

    @property
    def ymax(self):
        """Returns the maximum y-value of the bounding box."""
        return self.bbox[1][1]

    @ymax.setter
    def ymax(self, destination) -> None:
        """Sets the y-coordinate of the maximum edge of the bounding box.

        Args:
            destination : int or float y-coordinate of the maximum edge of the bbox.
        """
        self.move(destination=(0, destination), origin=self.bbox[1], axis="y")

    @property
    def xmin(self):
        """Returns the minimum x-value of the bounding box."""
        return self.bbox[0][0]

    @xmin.setter
    def xmin(self, destination) -> None:
        """Sets the x-coordinate of the minimum edge of the bounding box.

        Args:
            destination : int or float x-coordinate of the minimum edge of the bbox.
        """
        self.move(destination=(destination, 0), origin=self.bbox[0], axis="x")

    @property
    def ymin(self):
        """Returns the minimum y-value of the bounding box."""
        return self.bbox[0][1]

    @ymin.setter
    def ymin(self, destination) -> None:
        """Sets the y-coordinate of the minimum edge of the bounding box.

        Args:
            destination : int or float y-coordinate of the minimum edge of the bbox.
        """
        self.move(destination=(0, destination), origin=self.bbox[0], axis="y")

    @property
    def size(self):
        """Returns the (x, y) size of the bounding box."""
        bbox = self.bbox
        return bbox[1] - bbox[0]

    @property
    def xsize(self):
        """Returns the horizontal size of the bounding box."""
        bbox = self.bbox
        return bbox[1][0] - bbox[0][0]

    @property
    def ysize(self):
        """Returns the vertical size of the bounding box."""
        bbox = self.bbox
        return bbox[1][1] - bbox[0][1]

    def movex(self, origin=0, destination=None):
        """Moves an object by a specified x-distance.

        Args:
            origin: array-like[2], Port, or key Origin point of the move.
            destination: array-like[2], Port, key, or None Destination point of the move.
        """
        if destination is None:
            destination = origin
            origin = 0
        return self.move(origin=(origin, 0), destination=(destination, 0))

    def movey(self, origin=0, destination=None):
        """Moves an object by a specified y-distance.

        Args:
            origin : array-like[2], Port, or key Origin point of the move.
            destination : array-like[2], Port, or key Destination point of the move.
        """
        if destination is None:
            destination = origin
            origin = 0
        return self.move(origin=(0, origin), destination=(0, destination))

    def __add__(self, element) -> Group:
        """Adds an element to a Group.

        Args:
            element: Component, ComponentReference, Port, Polygon,
                Label, or Group to add.
        """
        if isinstance(self, Group):
            G = Group()
            G.add(self.elements)
            G.add(element)
        else:
            G = Group([self, element])
        return G


class Group(_GeometryHelper):
    """Group objects together so you can manipulate them as a single object \
            (move/rotate/mirror)."""

    def __init__(self, *args) -> None:
        """Initialize Group."""
        self.elements = []
        self.add(args)

    def __repr__(self) -> str:
        """Prints the number of elements in the Group."""
        return f"Group ({len(self.elements)} elements total)"

    def __len__(self) -> float:
        """Returns the number of elements in the Group."""
        return len(self.elements)

    def __iadd__(self, element) -> Group:
        """Adds an element to the Group.

        Args:
            element: Component, ComponentReference, Port, Polygon,
                Label, or Group to add.

        """
        return self.add(element)

    @property
    def bbox(self):
        """Returns the bounding boxes of the Group."""
        if len(self.elements) == 0:
            raise ValueError("Group is empty, no bbox is available")
        bboxes = np.empty([len(self.elements), 4])
        for n, e in enumerate(self.elements):
            bboxes[n] = e.bbox.flatten()

        bbox = (
            (bboxes[:, 0].min(), bboxes[:, 1].min()),
            (bboxes[:, 2].max(), bboxes[:, 3].max()),
        )
        return np.array(bbox)

    def add(self, element) -> Group:
        """Adds an element to the Group.

        Args:
            element: Component, ComponentReference, Port, Polygon,
                Label, or Group to add.
        """
        from gdsfactory.component import Component
        from gdsfactory.component_reference import ComponentReference

        if _is_iterable(element):
            [self.add(e) for e in element]
        elif element is None:
            return self
        elif isinstance(
            element, Component | ComponentReference | Polygon | Label | Group
        ):
            self.elements.append(element)
        else:
            raise ValueError(
                "add() Could not add element to Group, the only "
                "allowed element types are "
                "(Component, ComponentReference, Polygon, Label, Group)"
            )
        # Remove non-unique entries
        used = set()
        self.elements = [
            x for x in self.elements if x not in used and (used.add(x) or True)
        ]
        return self

    def rotate(self, angle: float = 45, center=(0, 0)) -> Group:
        """Rotates all elements in a Group around the specified centerpoint.

        Args:
            angle : int or float
                Angle to rotate the Group in degrees.
            center : array-like[2] or None
                center of the Group.
        """
        for e in self.elements:
            e.rotate(angle=angle, center=center)
        return self

    def move(self, origin=(0, 0), destination=None, axis=None) -> Group:
        """Moves the Group from the origin point to the destination.

        Both origin and destination can be 1x2 array-like, Port, or a key
        corresponding to one of the Ports in this Group.

        Args:
            origin : array-like[2], Port, or key
                Origin point of the move.
            destination : array-like[2], Port, or key
                Destination point of the move.
            axis : {'x', 'y'}
                Direction of the move.
        """
        destination = snap_to_grid(destination)
        for e in self.elements:
            e.move(origin=origin, destination=destination, axis=axis)
        return self

    def mirror(self, p1=(0, 1), p2=(0, 0)) -> Group:
        """Mirrors a Group across the line formed between the two specified points.

        ``points`` may be input as either single points
        [1,2] or array-like[N][2], and will return in kind.

        Args:
            p1 : array-like[N][2]
                First point of the line.
            p2 : array-like[N][2]
                Second point of the line.
        """
        for e in self.elements:
            e.mirror(p1=p1, p2=p2)
        return self

    def distribute(
        self, direction="x", spacing=100, separation=True, edge="center"
    ) -> Group:
        """Distributes the elements in the Group.

        Args:
            direction : {'x', 'y'}
                Direction of distribution; either a line in the x-direction or
                y-direction.
            spacing : int or float
                Distance between elements.
            separation : bool
                If True, guarantees elements are separated with a fixed spacing
                between; if False, elements are spaced evenly along a grid.
            edge : {'x', 'xmin', 'xmax', 'y', 'ymin', 'ymax'}
                Which edge to perform the distribution along (unused if
                separation == True)
        """
        _distribute(
            elements=self.elements,
            direction=direction,
            spacing=spacing,
            separation=separation,
            edge=edge,
        )
        return self

    def align(self, alignment="ymax") -> Group:
        """Aligns the elements in the Group.

        Args:
            alignment : {'x', 'y', 'xmin', 'xmax', 'ymin', 'ymax'}
                Which edge to align along (e.g. 'ymax' will align move the elements
                such that all of their topmost points are aligned)
        """
        _align(elements=self.elements, alignment=alignment)
        return self


def _rotate_points(points, angle: float = 45, center=(0, 0)):
    """Rotates points around a centerpoint defined by ``center``.

    ``points`` may be input as either single points [1,2] or array-like[N][2],
    and will return in kind.

    Args:
        points : array-like[N][2]
            Coordinates of the element to be rotated.
        angle : int or float
            Angle to rotate the points.
        center : array-like[2]
            Centerpoint of rotation.

    Returns:
        A new set of points that are rotated around ``center``.
    """
    if angle == 0:
        return points
    angle = angle * pi / 180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    c0 = np.array(center)
    if np.asarray(points).ndim == 2:
        return (points - c0) * ca + (points - c0)[:, ::-1] * sa + c0
    if np.asarray(points).ndim == 1:
        return (points - c0) * ca + (points - c0)[::-1] * sa + c0


def _reflect_points(points, p1=(0, 0), p2=(1, 0)):
    """Reflects points across the line formed by p1 and p2.

    from https://github.com/amccaugh/phidl/pull/181

    ``points`` may be input as either single points [1,2] or array-like[N][2],
    and will return in kind.

    Args:
        points : array-like[N][2]
            Coordinates of the element to be reflected.
        p1 : array-like[2]
            Coordinates of the start of the reflecting line.
        p2 : array-like[2]
            Coordinates of the end of the reflecting line.

    Returns:
        A new set of points that are reflected across ``p1`` and ``p2``.
    """
    original_shape = np.shape(points)
    points = np.atleast_2d(points)
    p1 = np.asarray(p1)
    p2 = np.asarray(p2)

    line_vec = p2 - p1
    line_vec_norm = norm(line_vec) ** 2

    # Compute reflection
    proj = np.sum(line_vec * (points - p1), axis=-1, keepdims=True)
    reflected_points = 2 * (p1 + (p2 - p1) * proj / line_vec_norm) - points

    return reflected_points if original_shape[0] > 1 else reflected_points[0]


def _is_iterable(items):
    """Checks if the passed variable is iterable.

    Args:
        items: any Item to check for iterability.
    """
    return isinstance(items, list | tuple | set | np.ndarray)


def _parse_coordinate(c):
    """Translates various inputs (lists, tuples, Ports) to an (x,y) coordinate.

    Args:
        c: array-like[N] or Port
            Input to translate into a coordinate.

    Returns:
        c : array-like[2]
            Parsed coordinate.
    """
    if hasattr(c, "center"):
        return c.center
    elif np.array(c).size == 2:
        return c
    else:
        raise ValueError(
            "Could not parse coordinate, input should be array-like (e.g. [1.5,2.3] or a Port"
        )


def _parse_move(origin, destination, axis):
    """Translates input coordinates to changes in position in the x and y directions.

    Args:
        origin : array-like[2] of int or float, Port, or key
            Origin point of the move.
        destination : array-like[2] of int or float, Port, key, or None
            Destination point of the move.
        axis : {'x', 'y'} Direction of move.

    Returns:
        dx : int or float
            Change in position in the x-direction.
        dy : int or float
            Change in position in the y-direction.
    """
    # If only one set of coordinates is defined, make sure it's used to move things
    if destination is None:
        destination = origin
        origin = [0, 0]

    d = _parse_coordinate(destination)
    o = _parse_coordinate(origin)
    if axis == "x":
        d = (d[0], o[1])
    if axis == "y":
        d = (o[0], d[1])
    dx, dy = np.array(d) - o

    return dx, dy


def _distribute(elements, direction="x", spacing=100, separation=True, edge=None):
    """Takes a list of elements and distributes them either equally along a \
    grid or with a fixed spacing between them.

    Args:
        elements: array-like of gdsfactory objects
            Elements to distribute.
        direction: {'x', 'y'}
            Direction of distribution; either a line in the x-direction or
            y-direction.
        spacing: int or float
            Distance between elements.
        separation: bool
            If True, guarantees elements are separated with a fixed spacing between;
            if False, elements are spaced evenly along a grid.
        edge: {'x', 'xmin', 'xmax', 'y', 'ymin', 'ymax'}
            Which edge to perform the distribution along (unused if
            separation == True)

    Returns:
        elements : Component, ComponentReference, Port, Polygon, Label, or Group
            Distributed elements.
    """
    if len(elements) == 0:
        return elements
    if direction not in ({"x", "y"}):
        raise ValueError("distribute(): 'direction' argument must be either 'x' or'y'")
    if (
        (direction == "x")
        and (edge not in ({"x", "xmin", "xmax"}))
        and (not separation)
    ):
        raise ValueError(
            "distribute(): When `separation` == False and direction == 'x',"
            " the `edge` argument must be one of {'x', 'xmin', 'xmax'}"
        )
    if (
        (direction == "y")
        and (edge not in ({"y", "ymin", "ymax"}))
        and (not separation)
    ):
        raise ValueError(
            "distribute(): When `separation` == False and direction == 'y',"
            " the `edge` argument must be one of {'y', 'ymin', 'ymax'}"
        )

    if direction == "y":
        sizes = [e.ysize for e in elements]
    if direction == "x":
        sizes = [e.xsize for e in elements]

    spacing = np.array([spacing] * len(elements))

    if separation:  # Then `edge` doesn't apply
        if direction == "x":
            edge = "xmin"
        if direction == "y":
            edge = "ymin"
    else:
        sizes = np.zeros(len(spacing))

    # Calculate new positions and move each element
    start = elements[0].__getattribute__(edge)
    positions = np.cumsum(np.concatenate(([start], (spacing + sizes))))
    for n, e in enumerate(elements):
        e.__setattr__(edge, positions[n])
    return elements


def _align(elements, alignment="ymax"):
    """Aligns lists of gdsfactory elements.

    Args:
        elements : array-like of gdsfactory objects
            Elements to align.
        alignment : {'x', 'y', 'xmin', 'xmax', 'ymin', 'ymax'}
            Which edge to align along (e.g. 'ymax' will align move the elements such
            that all of their topmost points are aligned)


    Returns
        elements : array-like of gdsfactory objects
            Aligned elements.
    """
    if len(elements) == 0:
        return elements
    if alignment not in (["x", "y", "xmin", "xmax", "ymin", "ymax"]):
        raise ValueError(
            "'alignment' argument must be one of 'x','y','xmin', 'xmax', 'ymin','ymax'"
        )
    value = Group(elements).__getattribute__(alignment)
    for e in elements:
        e.__setattr__(alignment, value)
    return elements


def _line_distances(points, start, end):
    if np.all(start == end):
        return np.linalg.norm(points - start, axis=1)

    vec = end - start
    cross = np.cross(vec, start - points)
    return np.divide(abs(cross), np.linalg.norm(vec))


def _simplify(points, tolerance=0):
    """Ramer–Douglas–Peucker algorithm for line simplification.

    Takes an array of points of shape (N,2) and removes excess points in the line.
    The remaining points form a identical line to within `tolerance` from the original
    """
    # From https://github.com/fhirschmann/rdp/issues/7
    # originally written by Kirill Konevets https://github.com/kkonevets

    M = np.asarray(points)
    start, end = M[0], M[-1]
    dists = _line_distances(M, start, end)

    index = np.argmax(dists)
    dmax = dists[index]

    if dmax <= tolerance:
        return np.array([start, end])

    result1 = _simplify(M[: index + 1], tolerance)
    result2 = _simplify(M[index:], tolerance)

    return np.vstack((result1[:-1], result2))


if __name__ == "__main__":
    import gdsfactory as gf

    # c = gf.Component()
    # label = c.add_label("hi")
    # print(c.labels[0])
    # _demo()
    # s = Step()

    c = gf.Component("bend")
    b = c << gf.components.bend_circular(angle=30)
    s = c << gf.components.straight(length=5)
    s.connect("o1", b.ports["o2"])
    p = c.get_polygons(as_shapely_merged=True)
    c2 = gf.Component()
    c2.add_polygon(p, layer=(1, 0))
    c2.show()



from __future__ import annotations

import pandas as pd

import gdsfactory as gf


def sample_reticle(grid: bool = True, **kwargs) -> gf.Component:
    """Returns MZI with TE grating couplers.

    Args:
        grid: if True returns a grid of components.
        kwargs: passed to pack or grid.

    """
    test_info_mzi_heaters = dict(
        doe="mzis_heaters",
        analysis="mzi_heater",
        measurement="optical_loopback4_heater_sweep",
    )
    test_info_ring_heaters = dict(
        doe="ring_heaters",
        analysis="ring_heater",
        measurement="optical_loopback2_heater_sweep",
    )

    mzis = [
        gf.components.mzi2x2_2x2_phase_shifter(length_x=lengths)
        for lengths in [100, 200, 300]
    ]
    rings = [
        gf.components.ring_single_heater(length_x=length_x) for length_x in [10, 20, 30]
    ]

    spirals_sc = [
        gf.components.spiral_inner_io_fiber_array(
            length=length,
            info=dict(
                doe="spirals_sc",
                measurement="optical_loopback4",
                analysis="optical_loopback4_spirals",
            ),
            decorator=gf.labels.add_label_json,
        )
        for length in [20e3, 40e3, 60e3]
    ]

    mzis_te = [
        gf.components.add_fiber_array_optical_south_electrical_north(
            mzi,
            electrical_port_names=["top_l_e2", "top_r_e2"],
            **test_info_mzi_heaters,
        )
        for mzi in mzis
    ]
    rings_te = [
        gf.components.add_fiber_array_optical_south_electrical_north(
            ring,
            electrical_port_names=["l_e2", "r_e2"],
            **test_info_ring_heaters,
        )
        for ring in rings
    ]

    components = mzis_te + rings_te + spirals_sc

    if grid:
        return gf.grid(components, **kwargs)
    c = gf.pack(components, **kwargs)
    if len(c) > 1:
        raise ValueError(f"failed to pack into single group. Made {len(c)} groups.")
    return c[0]


if __name__ == "__main__":
    c = sample_reticle(grid=False)
    gdspath = c.write_gds("mask.gds", with_metadata=True)
    csvpath = gf.labels.write_labels.write_labels_gdstk(
        gdspath, prefixes=["{"], layer_label="TEXT"
    )

    df = pd.read_csv(csvpath)
    print(df)
    c.show(show_ports=True)



from __future__ import annotations

from functools import partial

import gdstk
import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.typings import LayerSpec


@cell
def regular_polygon(
    sides: int = 6,
    side_length: float = 10,
    layer: LayerSpec = "WG",
    port_type: str | None = "placement",
) -> Component:
    """Returns a regular N-sided polygon, with ports on each edge.

    Args:
        sides: number of sides for the polygon.
        side_length: of the edges.
        layer: Specific layer to put polygon geometry on.
        port_type: optical, electrical.
    """
    c = Component()
    polygon = gdstk.regular_polygon((0, 0), side_length, sides)
    c.add_polygon(polygon, layer=layer)
    a = side_length / (2 * np.tan(np.pi / sides))

    if port_type:
        for side_index in range(sides):
            angle = 270 + side_index * 360 / sides
            center = (a * np.cos(np.radians(angle)), a * np.sin(np.radians(angle)))
            c.add_port(
                name=f"o{side_index+1}",
                center=center,
                width=side_length,
                layer=layer,
                port_type=port_type,
                orientation=angle,
            )

    c.auto_rename_ports()
    return c


hexagon = partial(regular_polygon, sides=6)
octagon = partial(regular_polygon, sides=8)


if __name__ == "__main__":
    # c = regular_polygon(sides=8, side_length=20)
    # c = rectangle(size=(3, 2), centered=True, layer=(2, 3))
    c = octagon(side_length=20)
    c.show(show_ports=True)



"""FabC example."""

from __future__ import annotations

import sys
from functools import partial

from pydantic import BaseModel

import gdsfactory as gf
from gdsfactory.add_pins import add_pins_inside1nm
from gdsfactory.cross_section import get_cross_sections, strip
from gdsfactory.port import select_ports
from gdsfactory.technology import LayerLevel, LayerStack
from gdsfactory.typings import Layer


class LayerMap(BaseModel):
    WG: Layer = (10, 1)
    WG_CLAD: Layer = (10, 2)
    WGN: Layer = (34, 0)
    WGN_CLAD: Layer = (36, 0)
    PIN: Layer = (1, 10)


LAYER = LayerMap()
WIDTH_NITRIDE_OBAND = 0.9
WIDTH_NITRIDE_CBAND = 1.0

select_ports_optical = partial(select_ports, layers_excluded=((100, 0),))


def get_layer_stack_fab_c(thickness: float = 350.0) -> LayerStack:
    """Returns generic LayerStack."""
    return LayerStack(
        layers=dict(
            wg=LayerLevel(
                layer=(1, 0),
                zmin=0.0,
                thickness=0.22,
            ),
            wgn=LayerLevel(
                layer=LAYER.WGN,
                zmin=0.22 + 0.1,
                thickness=0.4,
            ),
        )
    )


add_pins = partial(add_pins_inside1nm, pin_length=0.5)

# cross_sections

strip_nc = partial(
    strip,
    width=WIDTH_NITRIDE_CBAND,
    layer=LAYER.WGN,
    bbox_layers=[LAYER.WGN_CLAD],
    bbox_offsets=[3],
    add_pins_function_name="add_pins",
    add_pins_function_module="gdsfactory.samples.pdk.fab_c",
)
strip_no = partial(
    strip_nc,
    width=WIDTH_NITRIDE_OBAND,
)

xs_nc = strip_nc()
xs_no = strip_no()

# LEAF COMPONENTS have pins
bend_euler_nc = partial(gf.components.bend_euler, cross_section=xs_nc)
straight_nc = partial(gf.components.straight, cross_section=xs_nc)
bend_euler_o = partial(gf.components.bend_euler, cross_section=xs_no)
straight_o = partial(gf.components.straight, cross_section=xs_no)

mmi1x2_nc = partial(
    gf.components.mmi1x2,
    width=WIDTH_NITRIDE_CBAND,
    width_mmi=3,
    cross_section=xs_nc,
)
mmi1x2_no = partial(
    gf.components.mmi1x2,
    width=WIDTH_NITRIDE_OBAND,
    cross_section=xs_no,
)

gc_nc = partial(
    gf.components.grating_coupler_elliptical,
    grating_line_width=0.6,
    layer_slab=None,
    cross_section=xs_nc,
)

# HIERARCHICAL COMPONENTS made of leaf components

mzi_nc = partial(
    gf.components.mzi,
    cross_section=xs_nc,
    splitter=mmi1x2_nc,
    straight=straight_nc,
    bend=bend_euler_nc,
)
mzi_no = partial(
    gf.components.mzi,
    cross_section=xs_no,
    splitter=mmi1x2_no,
    straight=straight_o,
    bend=bend_euler_o,
)


cells = dict(
    mmi1x2_nc=mmi1x2_nc,
    mmi1x2_no=mmi1x2_no,
    bend_euler_nc=bend_euler_nc,
    straight_nc=straight_nc,
    mzi_nc=mzi_nc,
    mzi_no=mzi_no,
    gc_nc=gc_nc,
)


layer_stack = get_layer_stack_fab_c()
cross_sections = get_cross_sections(sys.modules[__name__])

pdk = gf.Pdk(
    name="fab_c_demopdk",
    cells=cells,
    cross_sections=cross_sections,
    layer_stack=layer_stack,
)


if __name__ == "__main__":
    # c2 = mmi1x2_nc(cache=False)
    # d2 = c2.to_dict()

    # from jsondiff import diff

    # d = diff(d1, d2)
    # c.show(show_ports=True)

    c = mzi_nc()
    c.show()



message = """
gdsfactory.plugins have been moved to gplugins

Make sure you have gplugins installed and use gplugins instead of gdsfactory.plugins

You can replace:
    import gdsfactory.plugins -> import gplugins

You can install gplugins with:
    pip install gplugins
"""

raise ImportError(message)



"""Wires for electrical manhattan routes."""

from __future__ import annotations

from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.typings import CrossSectionSpec

wire_straight = partial(straight, cross_section="xs_metal_routing")


@gf.cell
def wire_corner(
    cross_section: CrossSectionSpec = "xs_metal_routing",
) -> Component:
    """Returns 45 degrees electrical corner wire.

    Args:
        cross_section: spec.
        with_bbox: if True, includes the bbox layer and bbox offsets of the cross_section
    """
    x = gf.get_cross_section(cross_section)
    layer = x.layer
    width = x.width

    c = Component()
    a = width / 2
    xpts = [-a, a, a, -a]
    ypts = [-a, -a, a, a]

    c.add_polygon([xpts, ypts], layer=layer)

    c.add_port(
        name="e1",
        center=(-a, 0),
        width=width,
        orientation=180,
        layer=layer,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        center=(0, a),
        width=width,
        orientation=90,
        layer=layer,
        port_type="electrical",
    )
    c.info["length"] = width
    c.info["dy"] = width
    x.add_bbox(c)
    return c


@gf.cell
def wire_corner45(
    cross_section: CrossSectionSpec = "xs_metal_routing",
    radius: float = 10,
) -> Component:
    """Returns 90 degrees electrical corner wire.

    Args:
        cross_section: spec.
    """
    x = gf.get_cross_section(cross_section)
    layer = x.layer
    width = x.width
    radius = x.radius if radius is None else radius
    if radius is None:
        raise ValueError(
            "Radius needs to be specified in wire_corner45 or in the cross_section."
        )

    c = Component()

    a = width / 2

    xpts = [0, radius + a, radius + a, -np.sqrt(2) * width]
    ypts = [-a, radius, radius + np.sqrt(2) * width, -a]

    c.add_polygon([xpts, ypts], layer=layer)

    c.add_port(
        name="e1",
        center=(0, 0),
        width=width,
        orientation=180,
        layer=layer,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        center=(radius, radius),
        width=width,
        orientation=90,
        layer=layer,
        port_type="electrical",
    )
    c.info["length"] = np.sqrt(2) * radius
    return c


@gf.cell
def wire_corner_sections(
    cross_section: CrossSectionSpec = "xs_metal_routing",
) -> Component:
    """Returns 90 degrees electrical corner wire, where all cross_section sections properly represented.

    Works well with symmetric cross_sections, not quite ready for asymmetric.

    Args:
        cross_section: spec.
    """
    x = gf.get_cross_section(cross_section)

    xmin, ymax = x.get_xmin_xmax()

    main_section = x.sections[0]

    all_sections = [main_section]
    all_sections.extend(x.sections)

    c = Component()

    for section in all_sections:
        layer = section.layer
        width = section.width
        offset = section.offset
        b = width / 2

        xpts = [xmin, offset - b, offset - b, offset + b, offset + b, xmin]
        ypts = [
            -offset + b,
            -offset + b,
            ymax,
            ymax,
            -offset - b,
            -offset - b,
        ]

        c.add_polygon([xpts, ypts], layer=layer)

    c.add_port(
        name="e1",
        center=(xmin, -(xmin + ymax) / 2),
        orientation=180,
        cross_section=x,
        layer=x.layer,
    )
    c.add_port(
        name="e2",
        center=((xmin + ymax) / 2, ymax),
        orientation=90,
        cross_section=x,
        layer=x.layer,
    )
    c.info["length"] = ymax - xmin
    c.info["dy"] = ymax - xmin
    x.add_bbox(c)
    return c


if __name__ == "__main__":
    # c = wire_straight()

    # xsection = gf.cross_section.cross_section(
    #     width=1,
    #     offset=0,
    #     layer="M3",
    #     bbox_layers = ("M1", "M2"),
    #     bbox_offsets = (1, 2),
    # )

    # c = wire_corner(cross_section=xsection, with_bbox=True)
    # c.show(show_ports=True)
    # section1 = gf.cross_section.Section(width=0.4, layer="M1", offset=0.5)
    # section2 = gf.cross_section.Section(width=0.4, layer="M1", offset=-0.5)
    # xsection = gf.cross_section.cross_section(
    #     width=0.4, offset=0, layer="M1", sections=[section1, section2]
    # )
    from gdsfactory.cross_section import metal_slotted

    # # c = wire_corner_sections(cross_section=metal_slotted)
    # # c.show(show_ports=True)
    # # c.pprint_ports()
    # # c.pprint()

    metal_slotted_bbox = metal_slotted(
        bbox_layers=("M1", "M2"),
        bbox_offsets=(2, 4),
    )

    port1 = gf.Port(
        name="init",
        orientation=270,
        center=(0, 0),
        cross_section=gf.get_cross_section(metal_slotted),
    )
    port2 = gf.Port(
        name="final",
        orientation=0,
        center=(-100, -100),
        cross_section=gf.get_cross_section(metal_slotted),
    )

    c = gf.Component()
    route = gf.routing.get_route_electrical(
        input_port=port1,
        output_port=port2,
        bend=wire_corner_sections(cross_section=metal_slotted_bbox),
        cross_section=metal_slotted_bbox,
    )
    c.add(route.references)
    c.show()

    # print(yaml.dump(c.to_dict()))



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coh_rx_single_pol import coh_rx_single_pol
from gdsfactory.routing.get_route import get_route, get_route_from_waypoints
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@cell
def coh_rx_dual_pol(
    bend: ComponentSpec = bend_euler,
    cross_section: CrossSectionSpec = "xs_sc",
    lo_splitter: ComponentSpec = "mmi1x2",
    signal_splitter: ComponentSpec | None = None,
    spol_coh_rx: ComponentSpec = coh_rx_single_pol,
    single_pol_rx_spacing: float = 50.0,
    splitter_coh_rx_spacing: float = 40.0,
    lo_input_coupler: ComponentSpec | None = None,
    signal_input_coupler: ComponentSpec | None = None,
) -> Component:
    """Dual polarization coherent receiver.

    Args:
        bend: 90 degrees bend library.
        cross_section: for routing (splitter to mzms and mzms to combiners).
        lo_splitter: splitter function for the LO input.
        signal_splitter: splitter function for the signal input.
        spol_coh_rx: function generating a coherent rx for a single polarization.
        single_pol_rx_spacing: vertical spacing between each single polarization coherent receiver.
        splitter_coh_rx_spacing: horizontal spacing between the signal splitter and the single pol coh rxs.
        lo_input_coupler: Optional coupler to add before the LO splitter.
        signal_input_coupler: Optional coupler to add before the signal splitter.
    """
    bend_spec = bend
    bend = gf.get_component(bend, cross_section=cross_section)

    spol_coh_rx = gf.get_component(spol_coh_rx)

    # ----- Draw single pol coherent receivers -----

    c = Component()

    single_rx_1 = c << spol_coh_rx
    single_rx_2 = c << spol_coh_rx

    single_rx_1.mirror((1, 0))

    # Separate the two receivers
    single_rx_2.movey(single_rx_1.ymin - single_pol_rx_spacing - single_rx_2.ymax)

    # ------------ Signal splitters and input coupler ---------------

    if signal_splitter is not None:
        signal_splitter = gf.get_component(signal_splitter)
        signal_spl = c << signal_splitter
        signal_spl.xmax = single_rx_1.xmin - splitter_coh_rx_spacing
        signal_spl.y = (single_rx_1.y + single_rx_2.y) / 2

        route = get_route(
            signal_spl.ports["o2"],
            single_rx_1.ports["signal_in"],
            bend=bend_spec,
            cross_section=cross_section,
            with_sbend=False,
        )
        c.add(route.references)

        route = get_route(
            signal_spl.ports["o3"],
            single_rx_2.ports["signal_in"],
            bend=bend_spec,
            cross_section=cross_section,
            with_sbend=False,
        )
        c.add(route.references)

    if signal_input_coupler is not None:
        signal_coupler = gf.get_component(signal_input_coupler)
        signal_coup = c << signal_coupler
        signal_coup.mirror((0, 1))
        if signal_splitter is not None:
            signal_coup.connect("o1", signal_spl.ports["o1"])

        else:
            signal_coup.xmax = single_rx_1.xmin - splitter_coh_rx_spacing
            signal_coup.y = (single_rx_1.y + single_rx_2.y) / 2

            route = get_route(
                signal_coup.ports["o1"],
                single_rx_1.ports["signal_in"],
                bend=bend_spec,
                cross_section=cross_section,
                with_sbend=False,
            )
            c.add(route.references)

            route = get_route(
                signal_coup.ports["o2"],
                single_rx_2.ports["signal_in"],
                bend=bend_spec,
                cross_section=cross_section,
                with_sbend=False,
            )
            c.add(route.references)

    # ------------ LO splitter and input coupler ---------------

    lo_splitter = gf.get_component(lo_splitter)
    lo_split = c << lo_splitter

    if signal_input_coupler is not None:
        xlim = signal_coup.xmin
    elif signal_splitter is not None:
        xlim = signal_spl.xmin
    else:
        xlim = single_rx_1.xmin

    lo_split.xmax = xlim - splitter_coh_rx_spacing
    lo_split.y = (single_rx_1.y + single_rx_2.y) / 2

    p0x, p0y = lo_split.ports["o2"].center
    p1x, p1y = single_rx_1.ports["LO_in"].center

    route = get_route_from_waypoints(
        [
            (p0x, p0y),
            (p0x + splitter_coh_rx_spacing / 4, p0y),
            (p0x + splitter_coh_rx_spacing / 4, p1y),
            (p1x, p1y),
        ],
        bend=bend_spec,
        cross_section=cross_section,
    )
    c.add(route.references)

    p0x, p0y = lo_split.ports["o3"].center
    p1x, p1y = single_rx_2.ports["LO_in"].center

    route = get_route_from_waypoints(
        [
            (p0x, p0y),
            (p0x + splitter_coh_rx_spacing / 4, p0y),
            (p0x + splitter_coh_rx_spacing / 4, p1y),
            (p1x, p1y),
        ],
        bend=bend_spec,
        cross_section=cross_section,
    )
    c.add(route.references)

    if lo_input_coupler is not None:
        lo_coupler = gf.get_component(lo_input_coupler)
        lo_coup = c << lo_coupler
        lo_coup.connect("o1", lo_split.ports["o1"])

    c.add_ports(single_rx_1.get_ports_list(port_type="electrical"), prefix="pol1_")
    c.add_ports(single_rx_2.get_ports_list(port_type="electrical"), prefix="pol2_")
    c.add_ports(single_rx_1.get_ports_list(port_type="placement"), prefix="pol1_")
    c.add_ports(single_rx_2.get_ports_list(port_type="placement"), prefix="pol2_")
    c.add_ports(
        lo_split.get_ports_list(port_type="optical", orientation=180), prefix="lo_"
    )
    return c


if __name__ == "__main__":
    c = coh_rx_dual_pol()
    # c = coh_rx_single_pol()
    c.show(show_ports=True)



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.path import arc
from gdsfactory.typings import CrossSectionSpec, LayerSpec


@gf.cell
def bend_circular_heater(
    radius: float | None = None,
    angle: float = 90,
    npoints: int | None = None,
    heater_to_wg_distance: float = 1.2,
    heater_width: float = 0.5,
    layer_heater: LayerSpec = "HEATER",
    with_bbox: bool = True,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Creates an arc of arclength `theta` starting at angle `start_angle`.

    Args:
        radius: in um. Defaults to cross_section.radius.
        angle: angle of arc (degrees).
        npoints: Number of points used per 360 degrees.
        heater_to_wg_distance: in um.
        heater_width: in um.
        layer_heater: for heater.
        with_bbox: box in bbox_layers and bbox_offsets to avoid DRC sharp edges.
        cross_section: specification (CrossSection, string, CrossSectionFactory dict).
        kwargs: cross_section settings.
    """
    x = gf.get_cross_section(cross_section)
    radius = radius or x.radius
    width = x.width

    offset = heater_to_wg_distance + width / 2
    s1 = gf.Section(
        width=heater_width,
        offset=+offset,
        layer=layer_heater,
    )
    s2 = gf.Section(
        width=heater_width,
        offset=-offset,
        layer=layer_heater,
    )
    sections = list(x.sections) + [s1, s2]

    xs = x.copy(sections=sections)
    p = arc(radius=radius, angle=angle, npoints=npoints)

    c = Component()
    path = p.extrude(xs)
    ref = c << path
    c.add_ports(ref.ports)
    c.absorb(ref)
    c.info["length"] = np.round(p.length(), 3)
    c.info["dx"] = abs(p.points[0][0] - p.points[-1][0])
    c.info["dy"] = abs(p.points[0][0] - p.points[-1][0])

    x.validate_radius(radius)
    if with_bbox:
        top = None if int(angle) in {180, -180, -90} else 0
        bottom = 0 if int(angle) in {-90} else None
        x.add_bbox(c, top=top, bottom=bottom)
    return c


if __name__ == "__main__":
    c = bend_circular_heater(heater_width=1, cross_section="xs_rc")
    print(c.ports)
    c.show(show_ports=True)



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.geometry import boolean
from gdsfactory.typings import LayerSpec


@gf.cell
def rectangular_ring(
    enclosed_size=(4.0, 2.0),
    width: float = 0.5,
    layer: LayerSpec = "WG",
    centered: bool = False,
) -> Component:
    """Returns a Rectangular Ring

    Args:
        enclosed_size: (width, height) of the enclosed area.
        width: width of the ring.
        layer: Specific layer to put polygon geometry on.
        centered: True sets center to (0,0), False sets south-west to (0,0).
    """
    rect_in = rectangle(size=enclosed_size, centered=centered, layer=layer).ref()
    rect_out = rectangle(
        size=[dim + 2 * width for dim in enclosed_size], centered=centered, layer=layer
    )
    if not centered:
        rect_in.move((width, width))
    return boolean(A=rect_out, B=rect_in, operation="A-B", layer=layer)


if __name__ == "__main__":
    c = rectangular_ring(centered=True)
    c.show()



"""Straight waveguide."""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.cross_section import CrossSectionSpec


@gf.cell
def straight(
    length: float = 10.0,
    npoints: int = 2,
    layer: gf.typings.LayerSpec | None = None,
    width: float | None = None,
    add_pins: bool = True,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Returns a Straight waveguide.

    Args:
        length: straight length (um).
        npoints: number of points.
        layer: layer to use. Defaults to cross_section.layer.
        width: width to use. Defaults to cross_section.width.
        add_pins: add pins to the component.
        cross_section: specification (CrossSection, string or dict).

    .. code::

        o1 -------------- o2
                length
    """
    p = gf.path.straight(length=length, npoints=npoints)
    x = gf.get_cross_section(cross_section)

    if layer or width:
        x = x.copy(layer=layer or x.layer, width=width or x.width)

    c = Component()
    path = p.extrude(x)
    ref = c << path
    c.add_ports(ref.ports)

    x.add_bbox(c, right=0, left=0)
    if add_pins:
        x.add_pins(c)
    c.info["length"] = length
    c.info["width"] = x.sections[0].width
    c.info["cross_section"] = cross_section

    c.add_route_info(cross_section=x, length=length)
    c.absorb(ref)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    c = straight(cross_section="xs_rc")
    # c = straight()
    print(c.info)
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.array_component import array
from gdsfactory.components.straight import straight
from gdsfactory.port import auto_rename_ports
from gdsfactory.routing.sort_ports import sort_ports_x
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@cell
def array_with_fanout(
    component: ComponentSpec = "pad",
    columns: int = 3,
    pitch: float = 150.0,
    waveguide_pitch: float = 10.0,
    start_straight_length: float = 5.0,
    end_straight_length: float = 40.0,
    component_port_name: str = "e4",
    bend: ComponentSpec = "bend_euler",
    bend_port_name1: str | None = None,
    bend_port_name2: str | None = None,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Returns component array in X axis with west facing waveguides.

    Args:
        component: to replicate.
        columns: number of components.
        pitch: for waveguides.
        waveguide_pitch: for output waveguides.
        start_straight_length: length of the start of the straight.
        end_straight_length: length of the straight at the end.
        component_port_name: for fanout.
        bend: spec.
        bend_port_name1: optional port name.
        bend_port_name2: optional port name.
        cross_section: cross_section spec.
    """
    c = Component()
    component = gf.get_component(component)
    bend = gf.get_component(bend, cross_section=cross_section)

    bend_ports = bend.get_ports_list()
    bend_ports = sort_ports_x(bend_ports)
    bend_ports.reverse()
    bend_port_name1 = bend_port_name1 or bend_ports[0].name
    bend_port_name2 = bend_port_name2 or bend_ports[1].name

    for col in range(columns):
        ref = component.ref()
        ref.x = col * pitch
        c.add(ref)
        ylength = col * waveguide_pitch + start_straight_length
        xlength = col * pitch + end_straight_length
        straight_ref = c << straight(
            length=ylength,
            cross_section=cross_section,
        )
        port_s1, port_s2 = straight_ref.get_ports_list()

        straight_ref.connect(port_s2.name, ref.ports[component_port_name])

        bend_ref = c.add_ref(bend)
        bend_ref.connect(bend_port_name1, straight_ref.ports[port_s1.name])
        straightx_ref = c << straight(
            length=xlength,
            cross_section=cross_section,
        )
        straightx_ref.connect(port_s2.name, bend_ref.ports[bend_port_name2])
        c.add_port(f"W_{col}", port=straightx_ref.ports[port_s1.name])
    auto_rename_ports(c)
    return c


@cell
def array_with_fanout_2d(
    pitch: float = 150.0,
    pitch_x: float | None = None,
    pitch_y: float | None = None,
    columns: int = 3,
    rows: int = 2,
    **kwargs,
) -> Component:
    """Returns 2D array with fanout waveguides facing west.

    Args:
        pitch: 2D pitch.
        pitch_x: defaults to pitch.
        pitch_y: defaults to pitch.
        columns: number of columns.
        rows: number of rows.

    keyword args:
        component: to replicate.
        pitch: in um.
        waveguide_pitch: for fanout in um.
        start_straight_length: length of the start of the straight in um.
        end_straight_length: length of the straight at the end in um.
        radius: bend radius in um.
        cross_section: cross_section factory.
        component_port_name:
        bend_port_name1:
        bend_port_name2:
    """
    pitch_y = pitch_y or pitch
    pitch_x = pitch_x or pitch
    row = array_with_fanout(columns=columns, pitch=pitch_x, **kwargs)
    return array(component=row, rows=rows, columns=1, spacing=(0, pitch_y))


def test_array_with_fanout() -> None:
    c1 = array_with_fanout_2d(columns=2, rows=2)
    assert len(c1.ports) == 4


def test_array() -> None:
    c1 = array_with_fanout_2d(columns=2, rows=2)
    assert len(c1.ports) == 4


if __name__ == "__main__":
    # import gdsfactory as gf
    # c1 = gf.components.pad()
    # c2 = array(component=c1, pitch=150, columns=2)
    # print(c2.ports.keys())
    # c = array_with_fanout(
    #     columns=3,
    #     waveguide_pitch=20,
    #     bend=gf.components.wire_corner,
    #     cross_section='metal_routing',
    #     layer=(2, 0),
    #     width=10,
    #     radius=11,
    # )
    c = array_with_fanout_2d()
    c.show(show_ports=True)



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.typings import ComponentSpec, Float2, LayerSpec


@cell
def pad(
    size: str | Float2 = (100.0, 100.0),
    layer: LayerSpec = "MTOP",
    bbox_layers: tuple[LayerSpec, ...] | None = None,
    bbox_offsets: tuple[float, ...] | None = None,
    port_inclusion: float = 0,
    port_orientation: float | None = None,
) -> Component:
    """Returns rectangular pad with ports.

    Args:
        size: x, y size.
        layer: pad layer.
        bbox_layers: list of layers.
        bbox_offsets: Optional offsets for each layer with respect to size.
            positive grows, negative shrinks the size.
        port_inclusion: from edge.
        port_orientation: in degrees.
    """
    c = Component()
    layer = gf.get_layer(layer)
    size = gf.get_constant(size)
    rect = compass(
        size=size, layer=layer, port_inclusion=port_inclusion, port_type="electrical"
    )
    c_ref = c.add_ref(rect)
    c.add_ports(c_ref.ports)
    c.info["size"] = (float(size[0]), float(size[1]))
    c.info["layer"] = layer
    c.absorb(c_ref)

    if bbox_layers and bbox_offsets:
        sizes = []
        for cladding_offset in bbox_offsets:
            size = (size[0] + 2 * cladding_offset, size[1] + 2 * cladding_offset)
            sizes.append(size)

        for layer, size in zip(bbox_layers, sizes):
            ref = c.add_ref(
                compass(
                    size=size,
                    layer=layer,
                )
            )
            c.absorb(ref)

    width = size[1] if port_orientation in [0, 180] else size[0]

    c.add_port(
        name="pad",
        port_type="vertical_dc",
        layer=layer,
        center=(0, 0),
        orientation=port_orientation,
        width=width,
    )
    return c


pad_rectangular = partial(pad, size="pad_size")
pad_small = partial(pad, size=(80, 80))


@cell
def pad_array(
    pad: ComponentSpec = "pad",
    spacing: tuple[float, float] = (150.0, 150.0),
    columns: int = 6,
    rows: int = 1,
    orientation: float | None = 270,
) -> Component:
    """Returns 2D array of pads.

    Args:
        pad: pad element.
        spacing: x, y pitch.
        columns: number of columns.
        rows: number of rows.
        orientation: port orientation in deg. None for low speed DC ports.
    """
    c = Component()
    pad = gf.get_component(pad)
    size = pad.settings.full["size"]
    c.info["size"] = size

    c.add_array(pad, columns=columns, rows=rows, spacing=spacing)
    width = size[0] if orientation in [90, 270] else size[1]

    for col in range(columns):
        for row in range(rows):
            c.add_port(
                name=f"e{row+1}{col+1}",
                center=(col * spacing[0], row * spacing[1]),
                width=width,
                orientation=orientation,
                port_type="electrical",
                layer=pad.info["layer"],
            )
    return c


pad_array90 = partial(pad_array, orientation=90)
pad_array270 = partial(pad_array, orientation=270)

pad_array0 = partial(pad_array, orientation=0, columns=1, rows=3)
pad_array180 = partial(pad_array, orientation=180, columns=1, rows=3)


if __name__ == "__main__":
    # c = pad_rectangular()
    # c = pad()
    # c = pad(layer_to_inclusion={(3, 0): 10})
    # print(c.ports)
    # c = pad(width=10, height=10)
    # print(c.ports.keys())
    # c = pad_array90()
    c = pad_array0()
    # c = pad_array270()
    # c.pprint_ports()
    # c = pad_array_2d(cols=2, rows=3, port_names=("e2",))
    # c = pad_array(columns=2, rows=2, orientation=270)
    # c.auto_rename_ports()
    c.show(show_ports=True)



import pathlib
from typing import Any

from pydantic import BaseModel, model_validator

from gdsfactory.technology.layer_views import LayerViews

# required_layers = ["TEXT"]
required_layers = []


class LayerMap(BaseModel):
    """You will need to create a new LayerMap with your specific foundry layers."""

    model_config = {"frozen": True}

    @model_validator(mode="after")
    @classmethod
    def check_required_layers(cls, data: Any) -> Any:
        for layer in required_layers:
            if layer not in data.model_fields.keys():
                raise ValueError(f"{layer} layer must be defined")
        return data

    @model_validator(mode="after")
    @classmethod
    def check_all_layers_are_tuples_of_int(cls, data: Any) -> Any:
        for key, layer in data.model_fields.items():
            layer = layer.default
            if (
                not isinstance(layer, tuple)
                or len(layer) != 2
                or not all(isinstance(x, int) for x in layer)
            ):
                raise TypeError(f"{key} = {layer} must be a tuple of two integers")
        return data


def lyp_to_dataclass(lyp_filepath: str | pathlib.Path, overwrite: bool = True) -> str:
    """Returns python LayerMap script from a klayout layer properties file lyp."""
    filepathin = pathlib.Path(lyp_filepath)
    filepathout = filepathin.with_suffix(".py")

    if filepathout.exists() and not overwrite:
        raise FileExistsError(f"You can delete {filepathout}")

    script = """
from gdsfactory.typings import Layer
from gdsfactory.technology.layer_map import LayerMap


class LayerMapFab(LayerMap):
"""
    lys = LayerViews.from_lyp(filepathin)
    for layer_name, layer in sorted(lys.get_layer_views().items()):
        script += f"    {layer_name}: Layer = ({layer.layer[0]}, {layer.layer[1]})\n"

    script += """

LAYER = LayerMapFab()
"""

    filepathout.write_text(script)
    return script


if __name__ == "__main__":
    layers = LayerMap()
    # from gdsfactory.config import PATH

    # print(lyp_to_dataclass(PATH.klayout_lyp))



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.grating_coupler_elliptical import grating_coupler_elliptical
from gdsfactory.typings import ComponentSpec


@gf.cell
def grating_coupler_array(
    grating_coupler: ComponentSpec = grating_coupler_elliptical,
    pitch: float = 127.0,
    n: int = 6,
    port_name: str = "o1",
    rotation: int = 0,
    with_loopback: bool = False,
) -> Component:
    """Array of grating couplers.

    Args:
        grating_coupler: ComponentSpec.
        pitch: x spacing.
        n: number of grating couplers.
        port_name: port name.
        rotation: rotation angle for each reference.
        with_loopback: if True, adds a loopback between edge GCs. Only works for rotation = 90 for now.
    """
    c = Component()
    grating_coupler = gf.get_component(grating_coupler)

    for i in range(n):
        gc = c << grating_coupler
        gc.rotate(rotation)
        gc.x = i * pitch
        port_name_new = f"o{i}"
        c.add_port(port=gc.ports[port_name], name=port_name_new)

    if with_loopback:
        if rotation != 90:
            raise ValueError(
                "with_loopback is currently only programmed to work with rotation = 90"
            )
        if grating_coupler.get_ports_list()[0].cross_section is None:
            routing_xs = gf.cross_section.cross_section(
                layer=grating_coupler.get_ports_list()[0].layer,
                width=grating_coupler.get_ports_list()[0].width,
                radius=10,
            )
        else:
            routing_xs = gf.get_cross_section(
                grating_coupler.get_ports_list()[0].cross_section
            )
        radius = routing_xs.radius

        steps = (
            {"dy": -radius},
            {"dx": -gc.xsize / 2 - radius},
            {"dy": gc.ysize + 2 * radius},
            {"dx": c.xsize + 2 * radius},
            {"dy": -gc.ysize - 2 * radius},
            {"dx": -gc.xsize / 2 - radius},
        )

        route = gf.routing.get_route_from_steps(
            port1=c.ports["o0"],
            port2=c.ports[f"o{n-1}"],
            steps=steps,
            cross_section=routing_xs,
        )
        c.add(route.references)

        c.ports.pop("o0")
        c.ports.pop(f"o{n-1}")

    return c


if __name__ == "__main__":
    c = grating_coupler_array(rotation=90, with_loopback=True)
    c.show(show_ports=True)



"""
Based on Gerber file spec:
https://www.ucamco.com/files/downloads/file_en/456/gerber-layer-format-specification-revision-2022-02_en.pdf

See Also:
- https://github.com/opiopan/pcb-tools-extension
- https://github.com/jamesbowman/cuflow/blob/master/gerber.py
"""
from pathlib import Path
from typing import Literal

from pydantic import BaseModel, Field

from gdsfactory import Component


class GerberLayer(BaseModel):
    name: str
    function: list[str]
    polarity: Literal["Positive", "Negative"]


class GerberOptions(BaseModel):
    header: list[str] | None = None
    mode: Literal["mm", "in"] = "mm"
    resolution: float = 1e-6
    int_size: int = 4


# For generating a gerber job json file
class BoardOptions(BaseModel):
    size: tuple | None = (None,)
    n_layers: int = (2,)


resolutions = {1e-3: 3, 1e-4: 4, 1e-5: 5, 1e-6: 6}


def number(n) -> str:
    i = int(round(n * 10000))
    return "%07d" % i


def points(pp: list):
    out = ""
    d = "D02"
    for x, y in pp:
        out += f"X{number(x)}Y{number(y)}{d}*\n"
        d = "D01"
    return out


def rect(x0, y0, x1, y1):
    return "D10*\n" + points([(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)])


def linestring(pp):
    return "D10*\n" + points(pp)


def polygon(pp):
    return "G36*\n" + points(pp) + "G37*\n" + "\n"


def to_gerber(
    component: Component,
    dirpath: Path,
    layermap_to_gerber_layer: dict[tuple[int, int], GerberLayer],
    options: GerberOptions = Field(default_factory=dict),
) -> None:
    """Writes each layer to a different Gerber file.

    Args:
        component: to export.
        dirpath: directory path.
        layermap_to_gerber_layer: map of GDS layer to GerberLayer.
        options: to save.
            header: List[str] | None = None
            mode: Literal["mm", "in"] = "mm"
            resolution: float = 1e-6
            int_size: int = 4
    """
    for layer_tup, layer in layermap_to_gerber_layer.items():
        filename = (dirpath / layer.name.replace(" ", "_")).with_suffix(".gbr")
        with open(filename, "w+") as f:
            header = options.header or [
                "Gerber file generated by gdsfactory",
                f"Component: {component.name}",
            ]

            # Write file spec info
            f.write("%TF.FileFunction," + ",".join(layer.function) + "*%\n")
            f.write(f"%TF.FilePolarity,{layer.polarity}*%\n")

            digits = resolutions[options.resolution]
            f.write(f"%FSLA{options.int_size}{digits}Y{options.int_size}{digits}X*%\n")

            # Write header comments
            f.writelines([f"G04 {line}*\n" for line in header])

            # Setup units/mode
            units = options.mode.upper()
            f.write(f"%MO{units}*%\n")
            f.write("%LPD*%")

            f.write("G01*\n")

            # Aperture definition
            f.write("%ADD10C,0.050000*%\n")

            # Only supports polygons for now
            for poly in component.get_polygons(by_spec=layer_tup):
                f.write(polygon(poly))

            # File end
            f.write("M02*\n")


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.config import PATH
    from gdsfactory.technology import LayerLevel, LayerStack, LayerView, LayerViews
    from gdsfactory.typings import Layer

    class LayerMap(BaseModel):
        F_Cu: Layer = (1, 0)
        In1_Cu: Layer = (2, 0)
        In2_Cu: Layer = (3, 0)
        B_Cu: Layer = (4, 0)
        F_Silkscreen: Layer = (11, 0)
        F_Mask: Layer = (21, 0)
        B_Mask: Layer = (22, 0)
        Edge_Cuts: Layer = (31, 0)

        DEVREC: Layer = (68, 0)
        PORT: Layer = (1, 10)
        PORTE: Layer = (1, 11)

    LAYER = LayerMap()

    class PCBViews(LayerViews):
        F_Cu = LayerView(
            layer=LAYER.F_Cu,
            color="red",
        )
        In1_Cu = LayerView(
            layer=LAYER.In1_Cu,
            color="limegreen",
        )
        In2_Cu = LayerView(
            layer=LAYER.In2_Cu,
            color="goldenrod",
        )
        B_Cu = LayerView(
            layer=LAYER.B_Cu,
            color="blue",
        )
        F_Silkscreen = LayerView(
            layer=LAYER.F_Silkscreen,
            color="khaki",
        )
        F_Mask = LayerView(
            layer=LAYER.F_Mask,
            color="violet",
        )
        B_Mask = LayerView(
            layer=LAYER.B_Mask,
            color="aqua",
        )
        Edge_Cuts = LayerView(
            layer=LAYER.Edge_Cuts,
            color="gold",
        )

    LAYER_VIEWS = PCBViews()

    def get_pcb_layer_stack(
        copper_thickness: float = 0.035,
        core_thickness: float = 1,
    ):
        return LayerStack(
            layers=dict(
                top_cu=LayerLevel(
                    layer=LAYER.F_Cu,
                    thickness=copper_thickness,
                    zmin=0.0,
                    material="cu",
                ),
                inner1_cu=LayerLevel(
                    layer=LAYER.In1_Cu,
                    thickness=copper_thickness,
                    zmin=0.0,
                    material="cu",
                ),
                inner_core=LayerLevel(
                    layer=LAYER.Edge_Cuts,
                    thickness=core_thickness,
                    zmin=-core_thickness / 2,
                    material="fr4",
                ),
                inner2_cu=LayerLevel(
                    layer=LAYER.In2_Cu,
                    thickness=copper_thickness,
                    zmin=0.0,
                    material="cu",
                ),
                bottom_cu=LayerLevel(
                    layer=LAYER.B_Cu,
                    thickness=copper_thickness,
                    zmin=0.0,
                    material="cu",
                ),
            )
        )

    LAYER_STACK = get_pcb_layer_stack()

    layermap_to_gerber = {
        LAYER.F_Cu: GerberLayer(
            name="F_Cu", function=["Copper", "L1", "Top"], polarity="Positive"
        ),
        LAYER.B_Cu: GerberLayer(
            name="B_Cu", function=["Copper", "L2", "Bot"], polarity="Positive"
        ),
        LAYER.F_Silkscreen: GerberLayer(
            name="F_Silkscreen", function=["Legend", "Top"], polarity="Positive"
        ),
        LAYER.F_Mask: GerberLayer(
            name="F_Mask", function=["SolderMask", "Top"], polarity="Negative"
        ),
        LAYER.B_Mask: GerberLayer(
            name="B_Mask", function=["SolderMask", "Bot"], polarity="Negative"
        ),
        LAYER.Edge_Cuts: GerberLayer(
            name="Edge_Cuts", function=["Profile"], polarity="Positive"
        ),
    }

    # from gdsfactory.install import install_klayout_technology
    # from gdsfactory.technology.klayout_tech import KLayoutTechnology
    # tech_dir = (pathlib.Path(__file__) / "..").resolve() / "klayout"
    # pcb_tech = KLayoutTechnology(name='PCB', layer_views=PCBViews())
    # pcb_tech.technology.dbu = 1e-3
    # pcb_tech.export_technology_files(tech_dir=str(tech_dir))
    #
    # install_klayout_technology(tech_dir=tech_dir, tech_name="PCB")

    c = gf.components.text(layer=LAYER.F_Cu)
    c = LAYER_VIEWS.preview_layerset()

    gerber_path = PATH.repo / "extra" / "gerber"

    # This requires that the PCB technology (commented-out code above) is installed
    c.show(technology="PCB")

    to_gerber(
        c,
        dirpath=gerber_path,
        layermap_to_gerber_layer=layermap_to_gerber,
        options=GerberOptions(resolution=1e-6),
    )



"""Containers are Pcells that contain another Pcell."""

from __future__ import annotations

import sys

from gdsfactory.add_padding import add_padding_container
from gdsfactory.add_termination import add_termination
from gdsfactory.functions import rotate
from gdsfactory.get_factories import get_cells
from gdsfactory.routing import add_pads_bot, add_pads_top
from gdsfactory.routing.add_electrical_pads_shortest import add_electrical_pads_shortest
from gdsfactory.routing.add_electrical_pads_top import add_electrical_pads_top
from gdsfactory.routing.add_fiber_array import add_fiber_array
from gdsfactory.routing.add_fiber_single import add_fiber_single
from gdsfactory.routing.fanout2x2 import fanout2x2

__all__ = [
    "add_electrical_pads_shortest",
    "add_electrical_pads_top",
    "add_fiber_array",
    "add_fiber_single",
    "add_padding_container",
    "add_termination",
    "fanout2x2",
    "rotate",
    "add_pads_top",
    "add_pads_bot",
]

containers = get_cells(sys.modules[__name__])



import warnings

from gplugins.klayout.dataprep.regions import (
    Region,
    RegionCollection,
    boolean_not,
    boolean_or,
    copy,
    size,
)

message = """
gdsfactory.simulation have been moved to gplugins

Make sure you have gplugins installed and use gplugins instead of gdsfactory.simulation

You can replace:
    import gdsfactory.geometry.maskprep_flat as dp -> import gplugins.klayout.dataprep.regions as dp

You can install gplugins with:
    pip install gplugins
"""

warnings.warn(message)


__all__ = [
    "boolean_not",
    "boolean_or",
    "copy",
    "size",
    "Region",
    "RegionCollection",
]

if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.geometry.maskprep_flat as dp
    from gdsfactory.generic_tech.layer_map import LAYER as l

    c = gf.Component()
    ring = c << gf.components.coupler_ring()
    floorplan = c << gf.components.bbox(ring.bbox, layer=l.FLOORPLAN)
    c.write_gds("src.gds")

    d = dp.RegionCollection(filepath="src.gds", layermap=dict(l))
    fill_cell = d.get_fill(
        d.FLOORPLAN - d.WG, size=(0.1, 0.1), spacing=(0.1, 0.1), fill_layers=(l.WG,)
    )
    fill_cell.write("fill.gds")
    gf.show("fill.gds")



"""Converts CSV of test site labels into a CSV test manifest."""

import json
import pathlib

import numpy as np
import pandas as pd

import gdsfactory as gf
from gdsfactory.samples.sample_reticle import sample_reticle


def write_test_manifest(csvpath: str | pathlib.Path) -> pd.DataFrame:
    """Converts CSV of test site labels into a CSV test manifest."""
    df_in = pd.read_csv(csvpath)

    # Initialize an empty list to collect the rows
    rows = []
    columns = [
        "name",
        "xopt",
        "yopt",
        "xelec",
        "yelec",
        "measurement",
        "measurement_settings",
        "doe",
        "analysis",
        "analysis_settings",
    ]

    for _, label in df_in.iterrows():
        x, y, _orientation = label.x, label.y, label.rotation
        text = label.text
        d = json.loads(text)

        row = [
            d["name"] + f"_{int(x)}_{int(y)}",
            [np.round(i + x, 3) for i in d["xopt"]],
            [np.round(i + y, 3) for i in d["yopt"]],
            [np.round(i + x, 3) for i in d["xelec"]],
            [np.round(i + y, 3) for i in d["yelec"]],
            d["measurement"],
            d["measurement_settings"],
            d["doe"],
            d["analysis"],
            d["analysis_settings"],
        ]
        rows.append(row)

    return pd.DataFrame(
        rows,
        columns=columns,
    )


if __name__ == "__main__":
    c = sample_reticle(grid=False)
    c.show(show_ports=True)
    gdspath = c.write_gds()
    csvpath = gf.labels.write_labels.write_labels_gdstk(
        gdspath, prefixes=("{",), layer_label="TEXT"
    )
    df = write_test_manifest(csvpath)
    df.to_csv("test_manifest.csv")
    print(df)



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.components.bbox import bbox
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler import coupler
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def add_trenches(
    component: ComponentSpec = coupler,
    cross_section: CrossSectionSpec = "xs_rc_with_trenches",
    top: bool = True,
    bot: bool = True,
    right: bool = False,
    left: bool = False,
    **kwargs,
) -> gf.Component:
    """Return component with trenches.

    Args:
        component: component to add to the trenches.
        cross_section: spec (CrossSection, string or dict).
        top: add top trenches.
        bot: add bot trenches.
        right: add right trenches.
        left: add left trenches.
        kwargs: component settings.
    """
    c = gf.Component()
    component = gf.get_component(component, **kwargs)
    xs = gf.get_cross_section(cross_section)

    layer_trench = gf.get_layer(xs.info.get("layer_trench"))
    width_trench = xs.info.get("width_trench")

    top = width_trench if top else 0
    bot = width_trench if bot else 0
    left = width_trench if left else 0
    right = width_trench if right else 0

    core = component
    clad = bbox(
        core.bbox, layer=layer_trench, top=top, bottom=bot, left=left, right=right
    )
    ref = c << gf.geometry.boolean(clad, core, operation="not", layer=layer_trench)

    c.add_ports(component.ports, cross_section=xs)
    c.copy_child_info(component)
    c.absorb(ref)
    xs.add_pins(c)
    xs.add_bbox(c)

    return c


add_trenches90 = partial(
    add_trenches, component=bend_euler, top=False, bot=True, right=True, left=False
)

if __name__ == "__main__":
    c = add_trenches()
    c.show(show_ports=True)



def ensure_six_digit_hex_color(color: str | int) -> str:
    if isinstance(color, int):
        color = f"#{color:06x}"

    color = color.replace("0x", "#")

    # Convert shortened format to 6-digit (#0fc -> #00ffcc)
    if color[0] == "#" and len(color) == 4:
        color = "#" + "".join([2 * str(c) for c in color[1:]])
    return color



from __future__ import annotations

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import ErrorType
from gdsfactory.geometry.functions import angles_deg, curvature, path_length, snap_angle
from gdsfactory.typings import Coordinate, Coordinates, CrossSectionSpec


def bezier_curve(t: ndarray, control_points: Coordinates) -> ndarray:
    """Returns bezier coordinates.

    Args:
        t: 1D array of points varying between 0 and 1.
        control_points: for the bezier curve.
    """
    from scipy.special import binom

    xs = 0.0
    ys = 0.0
    n = len(control_points) - 1
    for k in range(n + 1):
        ank = binom(n, k) * (1 - t) ** (n - k) * t**k
        xs += ank * control_points[k][0]
        ys += ank * control_points[k][1]

    return np.column_stack([xs, ys])


@gf.cell
def bezier(
    control_points: Coordinates = ((0.0, 0.0), (5.0, 0.0), (5.0, 1.8), (10.0, 1.8)),
    npoints: int = 201,
    with_manhattan_facing_angles: bool = True,
    start_angle: int | None = None,
    end_angle: int | None = None,
    cross_section: CrossSectionSpec = "xs_sc",
    with_bbox: bool = True,
    add_pins: bool = True,
    bend_radius_error_type: ErrorType | None = None,
) -> Component:
    """Returns Bezier bend.

    Args:
        control_points: list of points.
        npoints: number of points varying between 0 and 1.
        with_manhattan_facing_angles: bool.
        start_angle: optional start angle in deg.
        end_angle: optional end angle in deg.
        cross_section: spec.
        with_bbox: box in bbox_layers and bbox_offsets to avoid DRC sharp edges.
        add_pins: add pins to the component.
    """
    xs = gf.get_cross_section(cross_section)
    t = np.linspace(0, 1, npoints)
    path_points = bezier_curve(t, control_points)
    path = gf.Path(path_points)

    if with_manhattan_facing_angles:
        path.start_angle = start_angle or snap_angle(path.start_angle)
        path.end_angle = end_angle or snap_angle(path.end_angle)

    c = Component()
    bend = path.extrude(xs)
    bend_ref = c << bend
    c.add_ports(bend_ref.ports)
    c.absorb(bend_ref)
    curv = curvature(path_points, t)
    length = gf.snap.snap_to_grid(path_length(path_points))
    if max(np.abs(curv)) == 0:
        min_bend_radius = np.inf
    else:
        min_bend_radius = gf.snap.snap_to_grid(1 / max(np.abs(curv)))

    c.info["length"] = length
    c.info["min_bend_radius"] = min_bend_radius
    c.info["start_angle"] = path.start_angle
    c.info["end_angle"] = path.end_angle

    xs.validate_radius(min_bend_radius, bend_radius_error_type)

    if with_bbox:
        xs.add_bbox(c)
    if add_pins:
        xs.add_pins(c)
    return c


def find_min_curv_bezier_control_points(
    start_point: ndarray,
    end_point: Coordinate,
    start_angle: float,
    end_angle: float,
    npoints: int = 201,
    alpha: float = 0.05,
    nb_pts: int = 2,
) -> Coordinates:
    """Returns bezier control points that minimize curvature.

    Args:
        start_point: start point.
        end_point: end point.
        start_angle: start angle in deg.
        end_angle: end angle in deg.
        npoints: number of points varying between 0 and 1.
        alpha: weight for angle mismatch.
        nb_pts: number of control points.
    """
    from scipy.optimize import minimize

    t = np.linspace(0, 1, npoints)

    def array_1d_to_cpts(a):
        xs = a[::2]
        ys = a[1::2]
        return list(zip(xs, ys))

    def objective_func(p):
        """minimize  max curvaturea and negligible start angle and end angle mismatch"""
        ps = array_1d_to_cpts(p)
        control_points = [start_point] + ps + [end_point]
        path_points = bezier_curve(t, control_points)

        max_curv = max(np.abs(curvature(path_points, t)))

        angles = angles_deg(path_points)
        dstart_angle = abs(angles[0] - start_angle)
        dend_angle = abs(angles[-2] - end_angle)
        angle_mismatch = dstart_angle + dend_angle
        return angle_mismatch * alpha + max_curv

    x0, y0 = start_point[0], start_point[1]
    xn, yn = end_point[0], end_point[1]

    initial_guess = []
    for i in range(nb_pts):
        x = (i + 1) * (x0 + xn) / nb_pts
        y = (i + 1) * (y0 + yn) / nb_pts
        initial_guess += [x, y]

    # initial_guess = [(x0 + xn) / 2, y0, (x0 + xn) / 2, yn]
    res = minimize(objective_func, initial_guess, method="Nelder-Mead")
    p = res.x
    return [tuple(start_point)] + array_1d_to_cpts(p) + [tuple(end_point)]


if __name__ == "__main__":
    control_points = ((0.0, 0.0), (5.0, 0.0), (5.0, 5.0), (10.0, 5.0))
    c = bezier(control_points=control_points)
    c.show(show_ports=True)



import gdstk
import numpy as np
from gdstk import Polygon

import gdsfactory as gf


def manhattanize_polygon(
    p: Polygon,
    target_step: float = 0.05,
) -> Polygon:
    """Return a Manhattanized version of the input polygon (where non-x and non-y parallel segments are decomposed into a staircase of small x and y-parallel segments)

    Implemented in pure Python, and hence only suited to small polygons.

    Args:
        p: input polygon.
        target_step: target staircase step size.

    Returns:
        manhattanized polygon

    .. plot::
      :include-source:

      import gdsfactory as gf
      c = gf.Component()

      poly = gdstk.rectangle((-2, -2), (2, 2))
      poly.rotate(np.pi / 4)
      poly.scale(1, 0.5)
      init_poly = c.add_polygon(poly, layer=1)
      final_poly = gf.geometry.manhattanize_polygon(poly)
      c.add_polygon(final_poly, layer=2)
      c.plot_matplotlib()

    """
    p_manhattan = []
    points = list(p.points)
    points.append(p.points[0])
    x_polarity = 0
    for pt1, pt2 in zip(points[:-1], points[1:]):
        dx = pt2[0] - pt1[0]
        dy = pt2[1] - pt1[1]
        if dx == 0 or dy == 0:
            p_manhattan.append(pt1)
            p_manhattan.append(pt2)
        else:
            num_x_steps = abs(int(np.ceil(dx / target_step)))
            num_y_steps = abs(int(np.ceil(dy / target_step)))
            num_steps = np.min([num_x_steps, num_y_steps]) + 1
            x_step = dx / num_steps
            y_step = dy / num_steps
            cur_x = pt1[0]
            cur_y = pt1[1]
            for _ in range(num_steps):
                if not x_polarity % 2:
                    p_manhattan.extend(
                        (
                            (cur_x + 0.5 * x_step, cur_y),
                            (cur_x + 0.5 * x_step, cur_y + y_step),
                            (cur_x + x_step, cur_y + y_step),
                        )
                    )
                else:
                    p_manhattan.extend(
                        (
                            (cur_x, cur_y + 0.5 * y_step),
                            (cur_x + x_step, cur_y + 0.5 * y_step),
                            (cur_x + x_step, cur_y + y_step),
                        )
                    )
                cur_x += x_step
                cur_y += y_step
            x_polarity += 1
    return gdstk.Polygon(p_manhattan)


def test_manhattanize() -> None:
    c = gf.Component("route")
    poly = gdstk.rectangle((-2, -2), (2, 2))
    poly.rotate(np.pi / 4)
    poly.scale(1, 0.5)
    init_poly = c.add_polygon(poly, layer=1)
    final_poly = c.add_polygon(manhattanize_polygon(poly, target_step=0.05), layer=2)

    assert len(init_poly.points) == 4
    assert len(final_poly.points) == 354


if __name__ == "__main__":
    c = gf.Component()

    poly = gdstk.rectangle((-2, -2), (2, 2))
    poly.rotate(np.pi / 4)
    poly.scale(1, 0.5)
    init_poly = c.add_polygon(poly, layer=1)
    final_poly = c.add_polygon(manhattanize_polygon(poly), layer=2)
    c.show()

    # test_manhattanize()



from __future__ import annotations

import copy

import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bezier import bezier
from gdsfactory.config import ErrorType
from gdsfactory.typings import CrossSectionSpec, Float2


@cell
def bend_s(
    size: Float2 = (11.0, 1.8),
    npoints: int = 99,
    cross_section: CrossSectionSpec = "xs_sc",
    with_bbox: bool = True,
    add_pins: bool = True,
    **kwargs,
) -> Component:
    """Return S bend with bezier curve.

    stores min_bend_radius property in self.info['min_bend_radius']
    min_bend_radius depends on height and length

    Args:
        size: in x and y direction.
        npoints: number of points.
        cross_section: spec.
        with_bbox: box in bbox_layers and bbox_offsets to avoid DRC sharp edges.
        add_pins: add pins to the component.

    Keyword Args:
    """
    c = Component()
    dx, dy = size

    bend = bezier(
        control_points=((0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy)),
        npoints=npoints,
        cross_section=cross_section,
        with_bbox=with_bbox,
        add_pins=add_pins,
        **kwargs,
    )
    bend_ref = c << bend
    c.add_ports(bend_ref.ports)
    c.copy_child_info(bend)
    return c


def get_min_sbend_size(
    size: Float2 = (None, 10.0),
    cross_section: CrossSectionSpec = "xs_sc",
    num_points: int = 100,
    **kwargs,
) -> float:
    """
    Returns the minimum sbend size to comply with bend radius requirements.

     Args:
        size: in x and y direction. One of them is None, which is the size we need to figure out.
        cross_section: spec.
        num_points: number of points to iterate over between max_size and 0.1 * max_size
        kwargs: cross_section settings.

    """

    cross_section_f = gf.get_cross_section(cross_section, **kwargs)

    if size[0] is None:
        ind = 0
        known_s = size[1]
    elif size[1] is None:
        ind = 1
        known_s = size[0]
    else:
        raise ValueError("One of the two elements in size has to be None")

    min_radius = cross_section_f.radius

    if min_radius is None:
        raise ValueError("The min radius for the specified layer is not known!")

    min_size = np.inf

    # Guess sizes, iterate over them until we cannot achieve the min radius
    # the max size corresponds to an ellipsoid
    max_size = 2.5 * np.sqrt(np.abs(min_radius * known_s))

    sizes = np.linspace(max_size, 0.1 * max_size, num_points)

    for i, s in enumerate(sizes):
        sz = copy.deepcopy(size)
        sz[ind] = s
        # print(sz)
        try:
            bend_s(
                size=sz,
                cross_section=cross_section,
                bend_radius_error_type=ErrorType.ERROR,
                **kwargs,
            )
            # print(c.info['min_bend_radius'])
            min_size = sizes[i]
        except ValueError:
            break

    return min_size


if __name__ == "__main__":
    c = bend_s()
    # c = bend_s(bbox_offsets=[0.5], bbox_layers=[(111, 0)], width=2)
    # c = bend_s(size=[10, 2.5])  # 10um bend radius
    # c = bend_s(size=[20, 3], cross_section="xs_rc")  # 10um bend radius
    # c.pprint()
    # c = bend_s_biased()
    # print(c.info["min_bend_radius"])
    c.show(show_ports=False)



"""Run GENERIC TECHs LVS.

Usage:
    run_lvs.py (--help| -h)
    run_lvs.py (--layout=<layout_path>) (--netlist=<netlist_path>) [--thr=<thr>] [--run_dir=<run_dir_path>] [--topcell=<topcell_name>] [--run_mode=<run_mode>] [--verbose] [--lvs_sub=<sub_name>] [--no_net_names] [--spice_comments] [--scale]

Options:
    --help -h                           Print this help message.
    --layout=<layout_path>              The input GDS file path.
    --netlist=<netlist_path>            The input netlist file path.
    --thr=<thr>                         The number of threads used in run.
    --run_dir=<run_dir_path>            Run directory to save all the results [default: pwd]
    --topcell=<topcell_name>            Topcell name to use.
    --run_mode=<run_mode>               Select klayout mode Allowed modes (flat , deep, tiling). [default: deep ]
    --lvs_sub=<sub_name>                Substrate name used in your design.
    --verbose                           Detailed rule execution log for debugging.
    --no_net_names                      Discard net names in extracted netlist.
    --spice_comments                    Enable netlist comments in extracted netlist.
    --scale                             Enable scale of 1e6 in extracted netlist.
"""

import logging
import os
from datetime import datetime
from subprocess import check_call

import klayout.db
from docopt import docopt


def check_klayout_version() -> None:
    """
    check_klayout_version checks klayout version and makes sure it would work with the LVS.
    """
    # ======= Checking Klayout version =======
    klayout_v_ = os.popen("klayout -b -v").read()
    klayout_v_ = klayout_v_.split("\n")[0]
    klayout_v_list = []

    if klayout_v_ == "":
        logging.error("Klayout is not found. Please make sure klayout is installed.")
        exit(1)
    else:
        klayout_v_list = [int(v) for v in klayout_v_.split(" ")[-1].split(".")]

    if len(klayout_v_list) < 1 or len(klayout_v_list) > 3:
        logging.error("Was not able to get klayout version properly.")
        exit(1)
    elif len(klayout_v_list) >= 2 or len(klayout_v_list) <= 3:
        if klayout_v_list[1] < 28:
            logging.error("Prerequisites at a minimum: KLayout 0.28.0")
            logging.error(
                "Using this klayout version has not been assesed in this development. Limits are unknown"
            )
            exit(1)

    logging.info(f"Your Klayout version is: {klayout_v_}")


def check_layout_type(layout_path):
    """
    check_layout_type checks if the layout provided is GDS or OAS. Otherwise, kill the process. We only support GDS or OAS now.

    Parameters
    ----------
    layout_path : string
        string that represent the path of the layout.

    Returns
    -------
    string
        string that represent full absolute layout path.
    """

    if not os.path.isfile(layout_path):
        logging.error(
            f"## GDS file path {layout_path} provided doesn't exist or not a file."
        )
        exit(1)

    if ".gds" not in layout_path and ".oas" not in layout_path:
        logging.error(
            f"## Layout {layout_path} is not in GDSII or OASIS format. Please use gds format."
        )
        exit(1)

    return os.path.abspath(layout_path)


def get_top_cell_names(gds_path):
    """
    get_top_cell_names get the top cell names from the GDS file.

    Parameters
    ----------
    gds_path : string
        Path to the target GDS file.

    Returns
    -------
    List of string
        Names of the top cell in the layout.
    """
    layout = klayout.db.Layout()
    layout.read(gds_path)
    top_cells = [t.name for t in layout.top_cells()]

    return top_cells


def get_run_top_cell_name(arguments, layout_path):
    """
    get_run_top_cell_name Get the top cell name to use for running. If it's provided by the user, we use the user input.
    If not, we get it from the GDS file.

    Parameters
    ----------
    arguments : dict
        Dictionary that holds the user inputs for the script generated by docopt.
    layout_path : string
        Path to the target layout.

    Returns
    -------
    string
        Name of the topcell to use in run.

    """

    if arguments["--topcell"]:
        topcell = arguments["--topcell"]
    else:
        layout_topcells = get_top_cell_names(layout_path)
        if len(layout_topcells) > 1:
            logging.error(
                "## Layout has multiple topcells. Please use --topcell to determine which topcell you want to run on."
            )
            exit(1)
        else:
            topcell = layout_topcells[0]

    return topcell


def generate_klayout_switches(arguments, layout_path, netlist_path):
    """
    parse_switches Function that parse all the args from input to prepare switches for LVS run.

    Parameters
    ----------
    arguments : dict
        Dictionary that holds the arguments used by user in the run command. This is generated by docopt library.
    layout_path : string
        Path to the layout file that we will run LVS on.
    netlist_path : string
        Path to the netlist file that we will run LVS on.

    Returns
    -------
    dict
        Dictionary that represent all run switches passed to klayout.
    """
    switches = dict()

    # No. of threads
    thrCount = 2 if arguments["--thr"] is None else int(arguments["--thr"])
    switches["thr"] = str(int(thrCount))

    if arguments["--run_mode"] in ["flat", "deep", "tiling"]:
        switches["run_mode"] = arguments["--run_mode"]
    else:
        logging.error("Allowed klayout modes are (flat , deep , tiling) only")
        exit()

    if arguments["--lvs_sub"]:
        switches["lvs_sub"] = arguments["--lvs_sub"]
    else:
        switches["lvs_sub"] = "generic_tech_gnd"

    if arguments["--verbose"]:
        switches["verbose"] = "true"
    else:
        switches["verbose"] = "false"

    if arguments["--no_net_names"]:
        switches["spice_net_names"] = "false"
    else:
        switches["spice_net_names"] = "true"

    if arguments["--spice_comments"]:
        switches["spice_comments"] = "true"
    else:
        switches["spice_comments"] = "false"

    if arguments["--scale"]:
        switches["scale"] = "true"
    else:
        switches["scale"] = "false"

    switches["topcell"] = get_run_top_cell_name(arguments, layout_path)
    switches["input"] = os.path.abspath(layout_path)
    switches["schematic"] = os.path.abspath(netlist_path)

    return switches


def build_switches_string(sws: dict):
    """
    build_switches_string Build switches string from dictionary.

    Parameters
    ----------
    sws : dict
        Dictionary that holds the Antenna switches.
    """
    return " ".join(f"-rd {k}={v}" for k, v in sws.items())


def check_lvs_results(results_db_files: list) -> None:
    """
    check_lvs_results Checks the results db generated from run and report at the end if the LVS run failed or passed.

    Parameters
    ----------
    results_db_files : list
        A list of strings that represent paths to results databases of all the LVS runs.
    """

    if len(results_db_files) < 1:
        logging.error("Klayout did not generate any db results. Please check run logs")
        exit(1)


def run_check(lvs_file: str, path: str, run_dir: str, sws: dict):
    """
    run_check run LVS check.

    Parameters
    ----------
    lvs_file : str
        String that has the file full path to run.
    path : str
        String that holds the full path of the layout.
    run_dir : str
        String that holds the full path of the run location.
    sws : dict
        Dictionary that holds all switches that needs to be passed to the antenna checks.

    Returns
    -------
    string
        string that represent the path to the results output database for this run.

    """

    logging.info(
        f'Running GENERIC TECH {lvs_file} checks on design {path} on cell {sws["topcell"]}'
    )

    layout_base_name = os.path.basename(path).split(".")[0]
    new_sws = sws.copy()
    report_path = os.path.join(run_dir, f"{layout_base_name}.lvsdb")
    ext_net_path = os.path.join(run_dir, f"{layout_base_name}.cir")
    new_sws["report"] = report_path
    new_sws["target_netlist"] = ext_net_path

    sws_str = build_switches_string(new_sws)

    run_str = f"klayout -b -r {lvs_file} {sws_str}"
    check_call(run_str, shell=True)

    return report_path


def main(lvs_run_dir: str, arguments: dict) -> None:
    """
    main function to run the LVS.

    Parameters
    ----------
    lvs_run_dir : str
        String with absolute path of the full run dir.
    arguments : dict
        Dictionary that holds the arguments used by user in the run command. This is generated by docopt library.
    """

    ## Check Klayout version
    check_klayout_version()

    ## Check layout file existence
    layout_path = arguments["--layout"]
    if not os.path.exists(arguments["--layout"]):
        logging.error(
            f"The input GDS file path {layout_path} doesn't exist, please recheck."
        )
        exit(1)

    ## Check layout type
    layout_path = check_layout_type(layout_path)

    # Check netlist file existence
    netlist_path = arguments["--netlist"]
    if not os.path.exists(arguments["--netlist"]):
        logging.error(
            f"The input netlist file path {netlist_path} doesn't exist, please recheck."
        )
        exit(1)

    lvs_rule_deck = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "generic_tech.lvs"
    )

    ## Get run switches
    switches = generate_klayout_switches(arguments, layout_path, netlist_path)

    ## Run LVS check
    res_db_files = run_check(lvs_rule_deck, layout_path, lvs_run_dir, switches)

    ## Check run
    check_lvs_results(res_db_files)


if __name__ == "__main__":
    # arguments
    arguments = docopt(__doc__, version="RUN LVS: 1.0")

    # logs format
    now_str = datetime.utcnow().strftime("lvs_run_%Y_%m_%d_%H_%M_%S")

    if (
        arguments["--run_dir"] == "pwd"
        or arguments["--run_dir"] == ""
        or arguments["--run_dir"] is None
    ):
        lvs_run_dir = os.path.join(os.path.abspath(os.getcwd()), now_str)
    else:
        lvs_run_dir = os.path.abspath(arguments["--run_dir"])

    os.makedirs(lvs_run_dir, exist_ok=True)

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[
            logging.FileHandler(os.path.join(lvs_run_dir, f"{now_str}.log")),
            logging.StreamHandler(),
        ],
        format="%(asctime)s | %(levelname)-7s | %(message)s",
        datefmt="%d-%b-%Y %H:%M:%S",
    )

    # Calling main function
    main(lvs_run_dir, arguments)



from warnings import warn

import numpy as np

from gdsfactory.component_reference import ComponentReference
from gdsfactory.config import CONF
from gdsfactory.port import Port
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import Route


def validate_connections(
    ports1: list[Port], ports2: list[Port], routes: list[Route]
) -> list[Route]:
    """
    Validates that a set of Routes indeed connects the port-pairs listed in ports1 and ports2. If the Routes form valid connections between ports1 and ports2, the original Routes will be returned. If not, a RouteWarning will be raised, and a set of error traces will be returned instead.

    Args:
        ports1: the list of starting ports.
        ports2: the list of ending ports.
        routes: the list of Route objects, purportedly between ports1 and ports2.

    Returns:
        A list of Routes. If the input routes are valid, they will be returned as-is. Otherwise, a list of error traces will be returned and a RouteWarning will be raised.
    """
    connections_expected = {_connection_tuple(p1, p2) for p1, p2 in zip(ports1, ports2)}

    for route in routes:
        connection = _connection_tuple(*route.ports)
        if connection not in connections_expected:
            return make_error_traces(
                ports1=ports1,
                ports2=ports2,
                message="Unable to route bundle! Please check the ordering of your ports to ensure it is a possible topology.",
            )
    return routes


def make_error_traces(
    ports1: list[Port], ports2: list[Port], message: str
) -> list[Route]:
    """
    Creates a set of error traces showing the intended connectivity between ports1 and ports2. The specified message will be included in the RouteWarning that is raised.

    Args:
        ports1: the list of starting ports.
        ports2: the list of ending ports.
        message: a message to include in the RouteWarning that is raised.

    Returns:
        A list of Routes (the error traces).
    """
    import gdsfactory as gf
    from gdsfactory.routing.manhattan import RouteWarning

    warn(message, RouteWarning)
    error_routes = []
    for port1, port2 in zip(ports1, ports2):
        path = gf.path.Path([port1.center, port2.center])
        error_component = gf.path.extrude(path, layer=CONF.layer_error_path, width=1)
        error_ref = ComponentReference(error_component)
        error_route = Route(
            references=[error_ref], ports=list(error_ref.ports.values()), length=np.nan
        )
        error_routes.append(error_route)
    return error_routes


def _connection_tuple(port1: Port, port2: Port) -> tuple:
    c1 = snap_to_grid(port1.center)
    c2 = snap_to_grid(port2.center)
    return (tuple(c1), tuple(c2))


def is_invalid_bundle_topology(ports1: list[Port], ports2: list[Port]) -> bool:
    """Returns True if the bundle is topologically unroutable without introducing crossings.

    Args:
        ports1: the starting ports of the bundle
        ports2: the ending ports of the bundle

    Returns:
        True if the bundle is unroutable. False otherwise.
    """
    # draw lines between ports1 and ports2
    # if no lines intersect and EITHER
    # A. all are within 90 degrees of port angle (non-negative dot product) or
    # B. all are further than 90 degrees of port angle (negative dot product)
    # OR all lines intersect and all ports1 > 90, ports2 < 90, or vice versa
    # the topology is valid
    # (actually, the bundle can contain 2 groups-- one of each, and still maintain valid, as long as there are no crossings between them)
    import shapely.geometry as sg
    from shapely import intersection_all

    # this is not really quite angle, but a threshold to check if dot products are effectively above/below zero, excluding numerical errors
    ANGLE_TOLERANCE = 1e-10

    if len(ports1) < 2:
        # if there's only one route, the bundle topology is always valid
        return False
    if any(p.orientation is None for p in ports1 + ports2):
        # don't check if the ports do not have orientation
        return False

    lines = [sg.LineString([p1.center, p2.center]) for p1, p2 in zip(ports1, ports2)]

    # Positive if BOTH ports are EITHER facing towards OR away from the vector of the outgoing line between them
    # Zero if either is orthogonal
    # Negative if one is facing and the other not
    ports_facing = []
    for p1, p2 in zip(ports1, ports2):
        dy_line = p2.center[1] - p1.center[1]
        dx_line = p2.center[0] - p1.center[0]

        dy_p1 = np.sin(np.deg2rad(p1.orientation))
        dx_p1 = np.cos(np.deg2rad(p1.orientation))

        dy_p2 = np.sin(np.deg2rad(p2.orientation))
        dx_p2 = np.cos(np.deg2rad(p2.orientation))
        dot1 = np.vdot([dx_line, dy_line], [dx_p1, dy_p1])
        dot2 = np.vdot([-dx_line, -dy_line], [dx_p2, dy_p2])
        both_facing = dot1 * dot2
        # print(both_facing)
        ports_facing.append(both_facing)

    intersections = intersection_all(lines)
    # print(intersections)
    if intersections.is_empty and all(s < -ANGLE_TOLERANCE for s in ports_facing):
        return True
    elif not intersections.is_empty and all(s > ANGLE_TOLERANCE for s in ports_facing):
        return True

    # NOTE: there are more complicated cases we are ignoring for now and giving "the benefit of the doubt"
    # i.e. if ports2 is perpendicular to ports1 and located somewhere laterally in between ports1
    # or some cases where ports are not properly ordered
    # for now we call these cases potentially valid, but we could be stricter in the future
    return False



"""Dummy fill to keep density constant using klayout."""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.typings import LayerSpec, LayerSpecs, PathType


def fill(
    gdspath,
    layer_to_fill: LayerSpec,
    layer_to_fill_margin: float = 0,
    layers_to_avoid: tuple[LayerSpec, float] | None = None,
    cell_name: str | None = None,
    fill_cell_name: str = "fill_cell",
    create_new_fill_cell: bool = False,
    include_original: bool = False,
    fill_layers: LayerSpecs | None = None,
    fill_size: tuple[float, float] = (10, 10),
    fill_spacing: tuple[float, float] = (20, 20),
    fill_name: str | None = None,
    gdspath_out: PathType | None = None,
) -> None:
    """Write gds file with fill.

    Args:
        gdspath: GDS input.
        layer_to_fill: Layer that defines the region to fill.
        layer_to_fill_margin: in um.
        layers_to_avoid: Layer to avoid to margin (((1, 0), 3.5), (LAYER.SLAB, 2)).
        cell_name: Optional cell to fill. Defaults to top cell.
        fill_cell_name: Optional cell name to use as fill.
        create_new_fill_cell: creates new fill cell, otherwise uses fill_cell_name from gdspath.
        include_original: True include original gdspath. False returns only fill.
        fill_layers: if create_new_fill_cell=True, defines new fill layers.
        fill_size: if create_new_fill_cell=True, defines new fill size.
        fill_spacing: fill pitch in x and y.
        fill_name: name of the cell containing all fill cells. Defaults to Cell_fill.
        gdspath_out: Optional GDS output. Defaults to input.
    """
    import kfactory as kf
    import klayout.db as kdb

    lib = kf.kcl
    lib.read(filename=str(gdspath))
    cell = lib[cell_name or 0]
    fill_name = fill_name or f"{cell.name}_fill"

    c = kf.KCell(fill_name)

    if create_new_fill_cell:
        if lib.has_cell(fill_cell_name):
            raise ValueError(f"{fill_cell_name!r} already in {str(gdspath)!r}")

        if not fill_layers:
            raise ValueError(
                "You need to pass fill_layers if create_new_fill_cell=True"
            )
        fill_cell = kf.KCell(fill_cell_name)
        for layer in fill_layers:
            layer = gf.get_layer(layer)
            layer = kf.kcl.layer(*layer)
            fill_cell << kf.cells.straight.straight(
                width=fill_size[0], length=fill_size[1], layer=layer
            )
    else:
        fill_cell = lib[fill_cell_name]

    fill_cell_index = fill_cell.cell_index()  # fill cell index
    fill_cell_box = fill_cell.bbox().enlarged(
        fill_spacing[0] / 2 * 1e3, fill_spacing[1] / 2 * 1e3
    )
    fill_margin = kf.kdb.Point(0, 0)

    layer_to_fill = gf.get_layer(layer_to_fill)
    layer_to_fill = cell.kcl.layer(*layer_to_fill)
    region = kdb.Region()
    region_avoid_all = kdb.Region()

    if layers_to_avoid:
        for layer, margin in layers_to_avoid:
            layer = gf.get_layer(layer)
            layer = kf.kcl.layer(*layer)
            region_avoid = kdb.Region()
            region_avoid.insert(
                cell.begin_shapes_rec(layer)
            ) if layer != layer_to_fill else None
            region_avoid = region_avoid.size(margin * 1e3)
            region_avoid_all = region_avoid_all + region_avoid

    region.insert(cell.begin_shapes_rec(layer_to_fill))
    region.size(-layer_to_fill_margin * 1e3)
    region_to_fill = region - region_avoid_all

    c.fill_region(
        region_to_fill,
        fill_cell_index,
        fill_cell_box,
        fill_margin,
    )

    if include_original:
        _ = c << cell

    gdspath_out = gdspath_out or gdspath
    c.write(str(gdspath_out))


@gf.cell
def cell_with_pad():
    c = gf.Component()
    c << gf.components.mzi(decorator=gf.add_padding)
    pad = c << gf.components.pad(size=(2, 2))
    pad.movey(10)
    return c


if __name__ == "__main__":
    c = cell_with_pad()
    c.show()
    gdspath = c.write_gds("mzi_fill.gds")

    use_fill_cell = True
    use_fill_cell = False
    spacing = 50

    if use_fill_cell:
        fill(
            gdspath,
            fill_layers=("WG",),
            layer_to_fill=(67, 0),
            layers_to_avoid=(((1, 0), 0), ((49, 0), 0)),
            # layers_to_avoid=(((1, 0), 0),),
            fill_cell_name="pad_size2__2",
            create_new_fill_cell=False,
            fill_spacing=(spacing, spacing),
            fill_size=(1, 1),
            include_original=True,
            layer_to_fill_margin=25,
        )
    else:
        fill(
            gdspath,
            fill_layers=("WG",),
            layer_to_fill=(67, 0),
            layers_to_avoid=(((1, 0), 0),),
            fill_cell_name="fill_cell",
            create_new_fill_cell=True,
            fill_spacing=(1, 1),
            fill_size=(1, 1),
            layer_to_fill_margin=25,
            include_original=True,
        )
    gf.show(gdspath)



"""Stream GDS to Klayout. Requires gdsfactory KLayout integration."""

from __future__ import annotations

import json
import os
import pathlib
import socket
from pathlib import Path

from gdsfactory import config


def show(
    gds_filename: Path | str,
    keep_position: bool = True,
    technology: str | None = None,
    port: int = 8082,
    delete: bool = False,
) -> None:
    """Show GDS in klayout.

    Args:
        gds_filename: to show.
        keep_position: keep position and active layers.
        technology: Name of the KLayout technology to use.
        port: klayout server port.
        delete: deletes file.
    """
    if not os.path.isfile(gds_filename):
        raise ValueError(f"{gds_filename} does not exist")

    gds_filename = pathlib.Path(gds_filename)

    data = {
        "gds": os.path.abspath(gds_filename),
        "keep_position": keep_position,
        "technology": technology,
    }
    data_string = json.dumps(data)
    try:
        conn = socket.create_connection(("127.0.0.1", port), timeout=0.5)
        data = data_string + "\n"
        enc_data = data.encode()
        conn.sendall(enc_data)
        conn.settimeout(5)
    except OSError:
        config.logger.warning(
            "Could not connect to klive server. Is klayout open and klive plugin installed?"
        )
    else:
        msg = ""
        try:
            msg = conn.recv(1024).decode("utf-8")
            config.logger.info(f"Message from klive: {msg}")
        except OSError:
            config.logger.warning("klive didn't send data, closing")
        finally:
            conn.close()

    if delete:
        Path(gds_filename).unlink()


if __name__ == "__main__":
    import gdsfactory as gf

    # c = gf.components.mzi()
    c = gf.components.straight(length=10)
    # gdspath = c.write_gds()
    # show(gdspath)
    c.show()



from __future__ import annotations

import gdsfactory as gf

Float2 = tuple[float, float]
Coordinate = tuple[Float2, Float2]


def bbox_to_points(
    bbox,
    top: float = 0,
    bottom: float = 0,
    left: float = 0,
    right: float = 0,
) -> list[list[float]]:
    (xmin, ymin), (xmax, ymax) = bbox
    xmin = float(xmin)
    xmax = float(xmax)
    ymin = float(ymin)
    ymax = float(ymax)
    return [
        [xmin - left, ymin - bottom],
        [xmax + right, ymin - bottom],
        [xmax + right, ymax + top],
        [xmin - left, ymax + top],
    ]


@gf.cell_without_validator
def bbox(
    bbox: tuple[Coordinate, Coordinate] = ((-1.0, -1.0), (3.0, 4.0)),
    layer: tuple[int, int] = (1, 0),
    top: float = 0,
    bottom: float = 0,
    left: float = 0,
    right: float = 0,
) -> gf.Component:
    """Returns bounding box rectangle from coordinates.

    Args:
        bbox: Coordinates of the box [(x1, y1), (x2, y2)].
        layer: for bbox.
        top: north offset.
        bottom: south offset.
        left: west offset.
        right: east offset.
    """
    D = gf.Component()
    (xmin, ymin), (xmax, ymax) = bbox
    points = [
        [xmin - left, ymin - bottom],
        [xmax + right, ymin - bottom],
        [xmax + right, ymax + top],
        [xmin - left, ymax + top],
    ]
    D.add_polygon(points, layer=layer)
    return D


if __name__ == "__main__":
    from gdsfactory.generic_tech import get_generic_pdk

    PDK = get_generic_pdk()
    PDK.activate()
    c = gf.Component()
    a = c << gf.components.L()
    c << bbox(bbox=a.bbox, top=10, left=5, right=-2)
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler90 import coupler90
from gdsfactory.components.coupler_straight import coupler_straight
from gdsfactory.components.straight import straight
from gdsfactory.typings import (
    ComponentFactory,
    ComponentSpec,
    Coordinates,
    CrossSectionSpec,
    LayerSpecs,
)


@gf.cell
def coupler_ring(
    gap: float = 0.2,
    radius: float = 5.0,
    length_x: float = 4.0,
    coupler90: ComponentFactory = coupler90,
    bend: ComponentSpec = bend_euler,
    coupler_straight: ComponentFactory = coupler_straight,
    cross_section: CrossSectionSpec = "xs_sc",
    cross_section_bend: CrossSectionSpec | None = None,
    length_extension: float = 3,
) -> Component:
    r"""Coupler for ring.

    Args:
        gap: spacing between parallel coupled straight waveguides.
        radius: of the bends.
        length_x: length of the parallel coupled straight waveguides.
        coupler90: straight coupled to a 90deg bend.
        bend: bend spec.
        coupler_straight: two parallel coupled straight waveguides.
        cross_section: cross_section spec.
        cross_section_bend: optional bend cross_section spec.
        length_extension: for the ports.

    .. code::

          o2            o3
           |             |
            \           /
             \         /
           ---=========---
        o1    length_x   o4

    """
    c = Component()
    gap = gf.snap.snap_to_grid(gap, grid_factor=2)
    xs = gf.get_cross_section(cross_section)
    xs_no_pins = xs.copy(add_pins_function_name=None)

    cross_section_bend = cross_section_bend or xs
    xs_bend = gf.get_cross_section(cross_section_bend)
    xs_bend = xs_bend.copy(radius=radius, add_pins_function_name=None)

    # define subcells
    coupler90_component = coupler90(
        gap=gap,
        radius=radius,
        bend=bend,
        cross_section=xs_no_pins,
        cross_section_bend=xs_bend,
    )
    coupler_straight_component = coupler_straight(
        gap=gap,
        length=length_x,
        cross_section=xs_no_pins,
    )

    # add references to subcells
    cbl = c << coupler90_component
    cbr = c << coupler90_component
    cs = coupler_straight_component.ref()

    if length_x > 0:
        c.add(cs)

    # connect references
    y = coupler90_component.y
    cs.connect(port="o4", destination=cbr.ports["o1"])
    cbl.mirror(p1=(0, y), p2=(1, y))
    cbl.connect(port="o2", destination=cs.ports["o2"])

    s = straight(length=length_extension, cross_section=xs_no_pins)

    s1 = c << s
    s2 = c << s

    s1.connect("o2", cbl["o4"])
    s2.connect("o1", cbr["o4"])

    c.add_port("o1", port=s1["o1"])
    c.add_port("o2", port=cbl["o3"])
    c.add_port("o3", port=cbr["o3"])
    c.add_port("o4", port=s2["o2"])

    c.add_ports(cbl.get_ports_list(port_type="electrical"), prefix="cbl")
    c.add_ports(cbr.get_ports_list(port_type="electrical"), prefix="cbr")
    c.auto_rename_ports()
    xs.add_pins(c)
    return c


@gf.cell
def coupler_ring_point(
    coupler_ring: ComponentFactory = coupler_ring,
    open_layers: LayerSpecs = None,
    open_sizes: Coordinates | None = None,
    **kwargs,
) -> Component:
    """Coupler ring that removes some layers at the coupling region.

    This allows short interaction lengths (point couplers).

    Args:
        coupler_ring: coupler_ring component to process.
        open_layers: layers to perform the boolean operations on.
        open_sizes: sizes of the boxes to use to remove layers, centered at bus center.

    Keyword Args:
        gap: spacing between parallel coupled straight waveguides.
        radius: of the bends.
        length_x: length of the parallel coupled straight waveguides.
        coupler90: straight coupled to a 90deg bend.
        bend: bend spec.
        coupler_straight: two parallel coupled straight waveguides.
        cross_section: cross_section spec.
        length_extension: for the ports.
    """
    c = gf.Component()

    coupler_ring_component = coupler_ring(**kwargs)
    open_layers = open_layers or []
    open_sizes = open_sizes or []

    open_layers_tuples = [gf.get_layer(open_layer) for open_layer in open_layers]
    untouched_layers = list(
        set(coupler_ring_component.get_layers()) - set(open_layers_tuples)
    )

    for layer, size in zip(open_layers, open_sizes):
        subcomponent = coupler_ring_component.extract(layers=[layer])
        rectangle = gf.components.rectangle(size=size, layer=layer, centered=True)
        _ = c << gf.geometry.boolean(subcomponent, rectangle, "A-B", layer=layer)

    coupler_ref = c << coupler_ring_component.extract(layers=untouched_layers)
    c.add_ports(coupler_ring_component.get_ports_list())
    c.absorb(coupler_ref)
    return c


if __name__ == "__main__":
    c = coupler_ring(cross_section_bend="xs_sc_heater_metal")
    c.show(show_ports=False)



"""Based on phidl.geometry."""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.geometry.boolean import boolean
from gdsfactory.typings import LayerSpec


@gf.cell
def invert(
    elements,
    border: float = 10.0,
    precision: float = 1e-4,
    layer: LayerSpec = (1, 0),
) -> Component:
    """Returns inverted version of input shapes with additional border around the edges.

    Args:
        elements : Component(/Reference), list of Component(/Reference), or Polygon \
                A Component containing the polygons to invert.
        border: Size of the border around the inverted shape (border value is the \
                distance from the edges of the boundary box defining the inverted \
                shape to the border, and is applied to all 4 sides of the shape).
        precision: Desired precision for rounding vertex coordinates.
        layer: Specific layer(s) to put polygon geometry on.

    Returns
        Component with inverted version of the input shape(s) and the border(s).

    .. plot::
      :include-source:

      import gdsfactory as gf

      e1 = gf.components.ellipse(radii=(6, 6))
      c = gf.geometry.invert(e1)
      c.plot_matplotlib()

    """
    Temp = Component()
    if type(elements) is not list:
        elements = [elements]
    for e in elements:
        if isinstance(e, Component):
            Temp.add_ref(e)
        else:
            Temp.add(e)
    gds_layer, gds_datatype = gf.get_layer(layer)

    # Build the rectangle around the Component D
    R = gf.components.rectangle(
        size=(Temp.xsize + 2 * border, Temp.ysize + 2 * border), centered=True
    ).ref()
    R.center = Temp.center
    return boolean(
        A=R,
        B=Temp,
        operation="A-B",
        precision=precision,
        layer=layer,
    )


def test_invert() -> None:
    e1 = gf.components.ellipse(radii=(6, 6))
    c = invert(e1)
    assert int(c.area()) == 910


if __name__ == "__main__":
    # test_invert()
    e1 = gf.components.ellipse(radii=(6, 6))
    c = invert(e1)
    c.show(show_ports=True)



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, LayerSpec


def _compute_parameters(xs_bend, wrap_angle_deg, radius):
    r_bend = xs_bend.radius
    theta = wrap_angle_deg / 2.0
    size_x, dy = r_bend * np.sin(theta * np.pi / 180), r_bend - r_bend * np.cos(
        theta * np.pi / 180
    )
    bus_length = max(4 * size_x, 2 * radius)
    return (r_bend, size_x, dy, bus_length)


def _generate_bends(c, r_bend, wrap_angle_deg, cross_section):
    if wrap_angle_deg != 0:
        input_arc = gf.path.arc(radius=r_bend, angle=-wrap_angle_deg / 2.0)
        bend_middle_arc = gf.path.arc(radius=r_bend, angle=-wrap_angle_deg)

        bend_input_output = input_arc.extrude(cross_section=cross_section)

        bend_input = c << bend_input_output
        bend_middle = c << bend_middle_arc.extrude(cross_section=cross_section)
        bend_middle.rotate(180 + wrap_angle_deg / 2.0, center=c.center)

        bend_input.connect("o2", bend_middle.ports["o2"])

        bend_output = c << bend_input_output
        bend_output.mirror()
        bend_output.connect("o2", bend_middle.ports["o1"])

        return (c, bend_input, bend_middle, bend_output)
    else:
        return (c, None, None, None)


def _generate_straights(c, bus_length, size_x, bend_input, bend_output, cross_section):
    straight_left = c << gf.components.straight(
        length=(bus_length - 4 * size_x) / 2.0, cross_section=cross_section
    )

    straight_right = c << gf.components.straight(
        length=(bus_length - 4 * size_x) / 2.0, cross_section=cross_section
    )

    if None not in (bend_input, bend_output):
        straight_left.connect("o2", bend_input.ports["o1"])

        straight_right.connect("o1", bend_output.ports["o1"])
    else:
        straight_left.connect("o2", straight_right.ports["o1"])

    return (c, straight_left, straight_right)


def _generate_circles(
    c, radius: float, xs, bend_middle, straight_left, r_bend, dy: float
):
    """Returns Component, circle and circle_cladding.

    Args:
        c: component.
        radius: in um.
        xs: cross_section:
        bend_middle: bend spec.
        straight_left: spec.
        r_bend: spec.
        dy: in um.
    """

    circle = c << gf.components.circle(radius=radius, layer=xs.layer)

    circle_cladding = None
    if bend_middle is not None:
        circle.move(
            origin=circle.center,
            destination=(
                (bend_middle.ports["o1"].x + bend_middle.ports["o2"].x) / 2.0,
                straight_left.ports["o2"].y - 2 * dy + r_bend,
            ),
        )
    else:
        circle.move(
            origin=circle.center,
            destination=(straight_left.ports["o2"].center + (0, r_bend),),
        )

    if circle_cladding:
        circle_cladding.move(origin=circle_cladding.center, destination=circle.center)

    return (c, circle, circle_cladding)


def _absorb(c, *refs):
    for ref in list(refs):
        if ref is not None:
            c.absorb(ref)
    return c


@gf.cell
def disk(
    radius: float = 10.0,
    gap: float = 0.2,
    wrap_angle_deg: float = 180.0,
    parity: int = 1,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Disk Resonator.

    Args:
       radius: disk resonator radius.
       gap: Distance between the bus straight and resonator.
       wrap_angle_deg: Angle in degrees between 0 and 180.
        determines how much the bus straight wraps along the resonator.
        0 corresponds to a straight bus straight.
        180 corresponds to a bus straight wrapped around half of the resonator.
       parity (1 or -1): 1, resonator left from bus straight, -1 resonator to the right.
       cross_section: cross_section spec.

    """
    if parity not in (1, -1):
        raise ValueError("parity must be 1 or -1")

    if wrap_angle_deg < 0.0 or wrap_angle_deg > 180.0:
        raise ValueError("wrap_angle_deg must be between 0.0 and 180.0")

    c = gf.Component()

    xs = gf.get_cross_section(cross_section=cross_section)
    radius_disk = radius
    radius = radius + xs.width / 2.0 + gap
    xs_bend = xs.copy(radius=radius)

    r_bend, size_x, dy, bus_length = _compute_parameters(
        xs_bend, wrap_angle_deg, radius
    )

    c, bend_input, bend_middle, bend_output = _generate_bends(
        c, r_bend, wrap_angle_deg, xs_bend
    )

    c, straight_left, straight_right = _generate_straights(
        c, bus_length, size_x, bend_input, bend_output, xs_bend
    )

    c, circle, circle_cladding = _generate_circles(
        c, radius_disk, xs, bend_middle, straight_left, r_bend, dy
    )

    c = _absorb(
        c,
        circle,
        circle_cladding,
        straight_left,
        straight_right,
        bend_input,
        bend_middle,
        bend_output,
    )

    c.add_port("o1", port=straight_left.ports["o1"], layer="PORT")
    c.add_port("o2", port=straight_right.ports["o2"])
    xs.add_bbox(c)
    if parity == -1:
        c = c.rotate(180)
    return c


@gf.cell
def disk_heater(
    radius: float = 10.0,
    gap: float = 0.2,
    wrap_angle_deg: float = 180.0,
    parity: int = 1,
    cross_section: CrossSectionSpec = "xs_sc",
    heater_layer: LayerSpec = "HEATER",
    via_stack: ComponentSpec = "via_stack_heater_mtop",
    heater_width: float = 5.0,
    heater_extent: float = 2.0,
    via_width: float = 10.0,
    port_orientation: float | None = 90,
) -> Component:
    """Disk Resonator with top metal heater.

    Args:
       radius: disk resonator radius.
       gap: Distance between the bus straight and resonator.
       wrap_angle_deg: Angle in degrees between 0 and 180.
        determines how much the bus straight wraps along the resonator.
        0 corresponds to a straight bus straight.
        180 corresponds to a bus straight wrapped around half of the resonator.
       parity (1 or -1): 1, resonator left from bus straight, -1 resonator to the right.
       cross_section: cross_section spec.
       heater_layer: layer of the heater.
       heater_width: width of the heater.
       heater_extent: length of heater beyond disk.
       via_width: size of the square via at the end of the heater.
       port_orientation: in degrees.
    """
    c = gf.Component()
    xs = gf.get_cross_section(cross_section=cross_section)

    disk_instance = c << disk(
        radius=radius,
        gap=gap,
        wrap_angle_deg=wrap_angle_deg,
        parity=parity,
        cross_section=cross_section,
    )

    dx = disk_instance.xmax - disk_instance.xmin
    dy = disk_instance.ymax - disk_instance.ymin
    heater = c << gf.get_component(
        gf.components.rectangle,
        size=(dx + 2 * heater_extent, heater_width),
        layer=heater_layer,
    )
    heater.x = disk_instance.x
    heater.y = dy / 2 + disk_instance.ymin + (xs.width + gap) / 2

    via = gf.get_component(via_stack, size=(via_width, via_width))
    c1 = c << via
    c2 = c << via
    c1.xmax = heater.xmin
    c1.y = heater.y
    c2.xmin = heater.xmax
    c2.y = heater.y
    c.add_ports(disk_instance.get_ports_list())
    c.add_ports(c1.get_ports_list(orientation=port_orientation), prefix="e1")
    c.add_ports(c2.get_ports_list(orientation=port_orientation), prefix="e2")
    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    # c = disk_heater(wrap_angle_deg=75)
    c = disk()
    c.show(show_ports=True)



"""Write FabC example PDK into uPDK YAML format."""

from __future__ import annotations

from gdsfactory.samples.pdk.fab_c import pdk

if __name__ == "__main__":
    yaml_pdk_decription = pdk.to_updk()
    print(yaml_pdk_decription)



"""pack a list of components into as few components as possible.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""

from __future__ import annotations

import warnings
from typing import Any

import numpy as np
from pydantic import validate_call

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.name import get_name_short
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import Anchor, ComponentSpec, Float2, Number


def _pack_single_bin(
    rect_dict: dict[int, tuple[Number, Number]],
    aspect_ratio: tuple[Number, Number],
    max_size: tuple[float, float],
    sort_by_area: bool,
    density: float,
) -> tuple[dict[int, tuple[Number, Number, Number, Number]], dict[Any, Any]]:
    """Packs a dict of rectangles {id:(w,h)} and tries to.

    Pack it into a bin as small as possible with aspect ratio `aspect_ratio`
    Will iteratively grow the bin size until everything fits or the bin size
    reaches `max_size`.

    Args:
        rect_dict: dict of rectangles {id: (w, h)} to pack.
        aspect_ratio: x, y.
        max_size: tuple of max X, Y size.
        sort_by_area: sorts components by area.
        density: of packing, closer to 1 packs tighter (more compute heavy).

    Returns:
        packed rectangles dict {id:(x,y,w,h)}. dict of remaining unpacked rectangles.

    """
    import rectpack

    # Compute total area and use it for an initial estimate of the bin size
    total_area = sum(r[0] * r[1] for r in rect_dict.values())
    aspect_ratio = np.asarray(aspect_ratio) / np.linalg.norm(aspect_ratio)  # Normalize

    # Setup variables
    box_size = np.asarray(aspect_ratio * np.sqrt(total_area), dtype=np.float64)
    box_size = np.clip(box_size, None, max_size)
    rp_sort = rectpack.SORT_AREA if sort_by_area else rectpack.SORT_NONE
    # Repeatedly run the rectangle-packing algorithm with increasingly larger
    # areas until everything fits or we've reached the maximum size
    while True:
        # Create the pack object
        rect_packer = rectpack.newPacker(
            mode=rectpack.PackingMode.Offline,
            pack_algo=rectpack.MaxRectsBlsf,
            sort_algo=rp_sort,
            bin_algo=rectpack.PackingBin.BBF,
            rotation=False,
        )

        # Add each rectangle to the pack, create a single bin, and pack
        for rid, r in rect_dict.items():
            rect_packer.add_rect(width=r[0], height=r[1], rid=rid)
        rect_packer.add_bin(width=box_size[0], height=box_size[1])
        rect_packer.pack()

        # Adjust the box size for next time
        box_size *= density  # Increase area to try to fit
        box_size = np.clip(box_size, None, max_size)

        # Quit the loop if we've packed all the rectangles or reached the max size
        if len(rect_packer.rect_list()) == len(rect_dict):
            break
        if all(box_size >= max_size):
            break

    # Separate packed from unpacked rectangles, make dicts of form {id:(x,y,w,h)}
    packed_rect_dict = {r[-1]: r[:-1] for r in rect_packer[0].rect_list()}
    unpacked_rect_dict = {
        k: v for k, v in rect_dict.items() if k not in packed_rect_dict
    }

    return packed_rect_dict, unpacked_rect_dict


@validate_call
def pack(
    component_list: list[ComponentSpec],
    spacing: float = 10.0,
    aspect_ratio: Float2 = (1.0, 1.0),
    max_size: tuple[float | None, float | None] = (None, None),
    sort_by_area: bool = True,
    density: float = 1.1,
    precision: float = 1e-2,
    text: ComponentSpec | None = None,
    text_prefix: str = "",
    text_mirror: bool = False,
    text_rotation: int = 0,
    text_offsets: tuple[Float2, ...] = ((0, 0),),
    text_anchors: tuple[Anchor, ...] = ("cc",),
    name_prefix: str | None = None,
    rotation: int = 0,
    h_mirror: bool = False,
    v_mirror: bool = False,
    add_ports_prefix: bool = True,
    add_ports_suffix: bool = False,
) -> list[Component]:
    """Pack a list of components into as few Components as possible.

    Args:
        component_list: list or tuple.
        spacing: Minimum distance between adjacent shapes.
        aspect_ratio: (width, height) ratio of the rectangular bin.
        max_size: Limits the size into which the shapes will be packed.
        sort_by_area: Pre-sorts the shapes by area.
        density: Values closer to 1 pack tighter but require more computation.
        precision: Desired precision for rounding vertex coordinates.
        text: Optional function to add text labels.
        text_prefix: for labels. For example. 'A' will produce 'A1', 'A2', ...
        text_mirror: if True mirrors text.
        text_rotation: Optional text rotation.
        text_offsets: relative to component size info anchor. Defaults to center.
        text_anchors: relative to component (ce cw nc ne nw sc se sw center cc).
        name_prefix: for each packed component (avoids the Unnamed cells warning). \
                Note that the suffix contains a uuid so the name will not be deterministic.
        rotation: optional component rotation in degrees.
        h_mirror: horizontal mirror in y axis (x, 1) (1, 0). This is the most common.
        v_mirror: vertical mirror using x axis (1, y) (0, y).
        add_ports_prefix: adds port names with prefix.
        add_ports_suffix: adds port names with suffix.

    .. plot::
        :include-source:

        import gdsfactory as gf
        from functools import partial

        components = [gf.components.triangle(x=i) for i in range(1, 10)]
        c = gf.pack(
            components,
            spacing=20.0,
            max_size=(100, 100),
            text=partial(gf.components.text, justify="center"),
            text_prefix="R",
            name_prefix="demo",
            text_anchors=["nc"],
            text_offsets=[(-10, 0)],
            v_mirror=True,
        )
        c[0].plot()

    """
    if density < 1.01:
        raise ValueError(
            "pack() `density` argument is too small. "
            "The density argument must be >= 1.01"
        )

    # Sanitize max_size variable
    max_size = [np.inf if v is None else v for v in max_size]
    max_size = np.asarray(max_size, dtype=np.float64)  # In case it's integers
    max_size = max_size / precision

    component_list = [gf.get_component(component) for component in component_list]

    # Convert Components to rectangles
    rect_dict = {}
    for n, D in enumerate(component_list):
        w, h = (D.size + spacing) / precision
        w, h = int(w), int(h)
        if (w > max_size[0]) or (h > max_size[1]):
            raise ValueError(
                f"pack() failed because Component {D.name!r} has x or y "
                "dimension larger than `max_size` and cannot be packed.\n"
                f"size = {w*precision, h*precision}, max_size = {max_size*precision}"
            )
        rect_dict[n] = (w, h)

    packed_list = []
    while rect_dict:
        (packed_rect_dict, rect_dict) = _pack_single_bin(
            rect_dict,
            aspect_ratio=aspect_ratio,
            max_size=max_size,
            sort_by_area=sort_by_area,
            density=density,
        )
        packed_list.append(packed_rect_dict)

    components_packed_list = []
    index = 0
    for i, rect_dict in enumerate(packed_list):
        name = get_name_short(f"{name_prefix or 'pack'}_{i}")
        packed = Component(name, with_uuid=True)
        packed.info["components"] = {}
        for n, rect in rect_dict.items():
            x, y, w, h = rect
            xcenter = x + w / 2 + spacing / 2
            ycenter = y + h / 2 + spacing / 2
            component = component_list[n]
            d = component.ref(rotation=rotation, h_mirror=h_mirror, v_mirror=v_mirror)
            packed.add(d)

            if hasattr(component, "settings"):
                packed.info["components"][component.name] = dict(component.settings)
            d.center = snap_to_grid((xcenter * precision, ycenter * precision))
            if add_ports_prefix:
                packed.add_ports(d.ports, prefix=f"{index}_")
            elif add_ports_suffix:
                packed.add_ports(d.ports, suffix=f"_{index}")
            else:
                try:
                    packed.add_ports(d.ports)
                except ValueError:
                    packed.add_ports(d.ports, suffix=f"_{index}")

            index += 1

            if text:
                for text_offset, text_anchor in zip(text_offsets, text_anchors):
                    label = packed << text(f"{text_prefix}{index}")
                    if text_mirror:
                        label.mirror()
                    if text_rotation:
                        label.rotate(text_rotation)
                    label.move(
                        np.array(text_offset) + getattr(d.size_info, text_anchor)
                    )

        components_packed_list.append(packed)

    if len(components_packed_list) > 1:
        groups = len(components_packed_list)
        warnings.warn(f"unable to pack in one component, creating {groups} components")

    return components_packed_list


def test_pack() -> None:
    """Test packing function."""
    component_list = [
        gf.components.ellipse(radii=tuple(np.random.rand(2) * n + 2)) for n in range(2)
    ]
    component_list += [
        gf.components.rectangle(size=tuple(np.random.rand(2) * n + 2)) for n in range(2)
    ]

    components_packed_list = pack(
        component_list,  # Must be a list or tuple of Components
        spacing=1.25,  # Minimum distance between adjacent shapes
        aspect_ratio=(2, 1),  # (width, height) ratio of the rectangular bin
        max_size=(None, None),  # Limits the size into which the shapes will be packed
        density=1.05,  # Values closer to 1 pack tighter but require more computation
        sort_by_area=True,  # Pre-sorts the shapes by area
    )
    c = components_packed_list[0]  # Only one bin was created, so we plot that
    assert len(c.get_dependencies()) == 4
    assert c


def test_pack_with_settings() -> None:
    """Test packing function with custom settings."""
    component_list = [
        gf.components.rectangle(size=(i, i), port_type=None) for i in range(1, 10)
    ]
    component_list += [
        gf.components.rectangle(size=(i, i), port_type=None) for i in range(1, 10)
    ]

    components_packed_list = pack(
        component_list,  # Must be a list or tuple of Components
        spacing=1.25,  # Minimum distance between adjacent shapes
        aspect_ratio=(2, 1),  # (width, height) ratio of the rectangular bin
        # max_size=(None, None),  # Limits the size into which the shapes will be packed
        max_size=(100, 100),  # Limits the size into which the shapes will be packed
        density=1.05,  # Values closer to 1 pack tighter but require more computation
        sort_by_area=True,  # Pre-sorts the shapes by area
        precision=1e-3,
    )
    assert components_packed_list[0]


if __name__ == "__main__":
    # test_pack()
    component_list = [
        gf.components.ellipse(radii=tuple(np.random.rand(2) * n + 2)) for n in range(2)
    ]
    component_list += [
        gf.components.rectangle(size=tuple(np.random.rand(2) * n + 2)) for n in range(2)
    ]

    components_packed_list = pack(
        component_list,  # Must be a list or tuple of Components
        spacing=1.25,  # Minimum distance between adjacent shapes
        aspect_ratio=(2, 1),  # (width, height) ratio of the rectangular bin
        max_size=(None, None),  # Limits the size into which the shapes will be packed
        density=1.05,  # Values closer to 1 pack tighter but require more computation
        sort_by_area=True,  # Pre-sorts the shapes by area
    )
    c = components_packed_list[0]  # Only one bin was created, so we plot that

    # p = pack(
    #     [gf.components.straight(length=i) for i in [1, 1]],
    #     spacing=20.0,
    #     max_size=(100, 100),
    #     text=partial(gf.components.text, justify="center"),
    #     text_prefix="R",
    #     name_prefix="demo",
    #     text_anchors=["nc"],
    #     text_offsets=[(-10, 0)],
    #     text_mirror=True,
    #     v_mirror=True,
    # )
    # c = p[0]
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler180
from gdsfactory.components.component_sequence import component_sequence
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def cutback_splitter(
    component: ComponentSpec = mmi1x2,
    cols: int = 4,
    rows: int = 5,
    port1: str = "o1",
    port2: str = "o2",
    port3: str = "o3",
    bend180: ComponentSpec = bend_euler180,
    mirror: bool = False,
    straight_length: float | None = None,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    """Returns a daisy chain of splitters for measuring their loss.

    Args:
        component: for cutback.
        cols: number of columns.
        rows: number of rows.
        port1: name of first optical port.
        port2: name of second optical port.
        bend180: ubend.
        straight: waveguide spec to connect both sides.
        mirror: Flips component. Useful when 'o2' is the port that you want to route to.
        straight_length: length of the straight section between cutbacks.
        cross_section: specification (CrossSection, string or dict).
        kwargs: cross_section settings.
    """
    xs = gf.get_cross_section(cross_section, **kwargs)

    component = gf.get_component(component)
    bendu = gf.get_component(bend180, cross_section=xs)
    straight_component = straight(
        length=straight_length or xs.radius * 2, cross_section=xs
    )

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "A": (component, port1, port2),
        "B": (component, port3, port1),
        "D": (bendu, "o1", "o2"),
        "C": (bendu, "o2", "o1"),
        "-": (straight_component, "o1", "o2"),
        "_": (straight_component, "o2", "o1"),
    }

    s = ""
    for i in range(rows):
        s += "AB" * cols
        if mirror:
            s += "C" if i % 2 == 0 else "D"
        else:
            s += "D" if i % 2 == 0 else "C"

    s = s[:-1]
    s += "-_"

    for i in range(rows):
        s += "AB" * cols
        s += "D" if (i + rows) % 2 == 0 else "C"

    s = s[:-1]

    seq = component_sequence(sequence=s, symbol_to_component=symbol_to_component)

    c = gf.Component()
    ref = c << seq
    c.add_ports(ref.ports)

    n = len(s) - 2
    c.copy_child_info(component)
    c.info["components"] = n
    return c


if __name__ == "__main__":
    c = cutback_splitter()
    c.show(show_ports=True)
